__version__ = "0.3.4"
__author__ = 'Tam V. Nguyen'
__credits__ = 'Helmholtz Centre for Environmental Research (UFZ)'

from .import data, interface, model, utility, train
__all__ = ["data", "interface", "model", "utility", "train"]


import tkinter as tk
import customtkinter as ctk
from hydroecolstm.model_run import run_config
from hydroecolstm.data.read_data import read_scale_data
from hydroecolstm.model.create_model import create_model
from hydroecolstm.train.trainer import Trainer
from hydroecolstm.model_run import config_to_search_space
from CTkToolTip import CTkToolTip
import torch
from PIL import Image
from pathlib import Path
from ray import tune    # is used within the eval function

class TrainTestFrame(ctk.CTkScrollableFrame):
    def __init__(self, container=None, config=None, globalData=None):
        super().__init__(container)
        
        # setup the grid layout manager
        self.config = config
        self.globalData = globalData
        self.__create_widgets() 
        
    # create widgets for sidebar frame 
    def __create_widgets(self): 
        
        # create tabs #
        self.tabview = ctk.CTkTabview(master=self, width = 750, border_width=1.5,
                                      fg_color = "transparent")
        self.tabview.pack(fill='both',expand=1)
        self.tabview.add("1. Initial state dicts")
        self.tabview.tab("1. Initial state dicts").grid_columnconfigure((0), weight=1)
        self.tabview.add("2. Trainer")
        self.tabview.tab("2. Trainer").grid_columnconfigure((0,1), weight=1)
        
        # Tooltip for tune seach space api
        self.ray_tooltip = ('Or input the tune Search Space API command, ' +
                            'e.g., tune.grid_search([20, 30]). Please see:' + 
                            'https://docs.ray.io/en/latest/tune/api/search_space.html')

        ctk.CTkButton(master=self,  anchor='w', fg_color='gray',
                      text="Parameter explanation",
                      command=self.__parameter_explanation).pack(anchor="e", pady = 10)
        # ---------------------------------------------Initial state dicts      
        # Load model state dict
        self.load_model_label = ctk.CTkLabel(self.tabview.tab("1. Initial state dicts"), 
                                      text="1. Load model state dicts")
        self.load_model_label.grid(row=0, column=0, sticky = "w")  
        self.load_model = ctk.CTkButton(self.tabview.tab("1. Initial state dicts"), anchor='w', 
                                         command=self.load_state_dict,
                                         text="Load model (optional)")
        self.load_model.grid(row=1, column=0, sticky = "w")
        CTkToolTip(self.load_model, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="Optional input - load initial model state dicts. " +
                   "This could be, e.g., the calibrated model at a regional scale " +
                   "and used in here for parameter fine tuning. The model " +
                   "state dict is ONLY USED with the MANUAL hyperparamter " +
                   "optimization, incase of AUTOMATIC hyperparameter tuning "+
                   "with ray tune, this option is discarded")
        
        # ---------------------------------------------Content of load data tab
        # Number of epochs
        self.nepoch_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="1. Number of epochs")
        self.nepoch_label.grid(row=0, column=0, sticky = "w")
        
        self.nepoch = ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="5")
               
        self.nepoch.grid(row=1, column=0, sticky = "w")
        self.nepoch.bind('<KeyRelease>', self.get_nepoch)
        CTkToolTip(self.nepoch, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message='Number of training epochs (positive integer number)') 
        
        # Learning rate
        self.learning_rate_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                                text="2. Learning rate")
        self.learning_rate_label.grid(row=2, column=0, sticky = "w")
        self.learning_rate= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="0.01")
        self.learning_rate.grid(row=3, column=0, sticky = "w")
        self.learning_rate.bind('<KeyRelease>', self.get_learning_rate)
        CTkToolTip(self.learning_rate, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message='Learning rate for gradient descent (positive real number)') 
        
        # Warm up length
        self.warmup_length_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="3. Warm-up length")
        self.warmup_length_label.grid(row=4, column=0, sticky = "w")
        self.warmup_length= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="30") 
        self.warmup_length.grid(row=5, column=0, sticky = "w")
        self.warmup_length.bind('<KeyRelease>', self.get_warmup_length)
        CTkToolTip(self.warmup_length, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message='The number of timesteps used for warm-up. \n' +
                   'For example, the first warmup_length outputs will be skipped \n'+
                   'when calculating the loss function. This value MUST \n'+
                   'be smaller than sequence length') 
        
        # Sequence length
        self.sequence_length_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="4. Sequence length")
        self.sequence_length_label.grid(row=6, column=0, sticky = "w")
        self.sequence_length= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="720") 
        self.sequence_length.grid(row=7, column=0, sticky = "w")
        self.sequence_length.bind('<KeyRelease>', self.get_sequence_length)
        CTkToolTip(self.sequence_length, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="The number of timesteps in each sample dataset. \n" + 
                   "One epoch loops over multiple baches. \n" +
                   "One batch consists of 'batch_size' sample datasets. \n" +
                   "One sample dataset consist of pairs of Input and Tartget \n" +
                   "    of 'sequence_length' timesteps (chronological order)")

        # Batch size
        self.batch_size_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="5. Batch size")
        self.batch_size_label.grid(row=8, column=0, sticky = "w")
        self.batch_size= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="3") 
        self.batch_size.grid(row=9, column=0, sticky = "w")
        self.batch_size.bind('<KeyRelease>', self.get_batch_size)
        CTkToolTip(self.batch_size, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="Please see sequence length for help") 
        
        # Patience
        self.patience_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                         text="6. Patience length")
        self.patience_label.grid(row=0, column=2, sticky = "w")
        self.patience= ctk.CTkEntry(master=self.tabview.tab("2. Trainer"),
                             placeholder_text="20") 
        self.patience.grid(row=1, column=2, sticky = "w")
        self.patience.bind('<KeyRelease>', self.get_patience_length)
        CTkToolTip(self.patience, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="The number of epochs to wait (before stopping) to \n" + 
                   "see if there is no improvement in ths validation loss, \n" + 
                   "which is used for early stopping. More information \n" + 
                   "please see https://github.com/Bjarten/early-stopping-pytorch. \n" +
                   "Patience length should be much smaller than the number of epochs")

        # Optimization method
        self.optim_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                        text="7. Optimization method")
        self.optim_label.grid(row=2, column=2, sticky = "w")    
        self.optim = ctk.CTkOptionMenu(self.tabview.tab("2. Trainer"),
                                                   values=['Adam'],
                                                   command=self.get_optim_method) 
        self.optim.grid(row=3, column=2, sticky = "w")
        CTkToolTip(self.optim, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="In this version only Adam method is available")
        
        # Loss function
        self.loss_function_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                               text="8. Loss function")
        self.loss_function_label.grid(row=4, column=2, sticky = "w")    
        self.loss = ctk.CTkOptionMenu(self.tabview.tab("2. Trainer"),
                                                   values=['Root Mean Square Error',
                                                           'Mean Absolute Error',
                                                           'Mean Squared Error',
                                                           '1 - Nash Sutcliffe'],
                                                   command=self.loss_function) 
        self.loss.grid(row=5, column=2, sticky = "w")
        CTkToolTip(self.loss, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="Please select the loss function for calculating"+
                   " traning and validatation losses, which are used for" +
                   " updating model paramters and early stopping")
        
        # Save model
        self.out_dir_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                      text="9. Output directory")
        self.out_dir_label.grid(row=6, column=2, sticky = "w")     
        self.out_dir = ctk.CTkButton(self.tabview.tab("2. Trainer"), anchor='w', 
                                         command=self.out_dir_event,
                                         text="Select directory")
        self.out_dir.grid(row=7, column=2, sticky = "w")
        CTkToolTip(self.out_dir, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="Please select the directory for saving model outputs"+
                   " If not selected, the default directory (pathlib.Path.cwd()) is used")
        
        
        # Run model
        self.run_label = ctk.CTkLabel(self.tabview.tab("2. Trainer"), 
                                      text="10. Run model")
        self.run_label.grid(row=8, column=2, sticky = "w")     
        self.run = ctk.CTkButton(self.tabview.tab("2. Trainer"), anchor='w', 
                                         command=self.run_train_test,
                                         text="Run")
        self.run.grid(row=9, column=2, sticky = "w")
        CTkToolTip(self.run, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="Click here to run the model. While model is running"+
                   " in the background, the GUI will be frozen.")
                
        # Progressbar
        self.progressbar = ctk.CTkProgressBar(master=self.tabview.tab("2. Trainer"))
        self.progressbar.grid(row=10, column=2,  sticky = "w", pady = (10,10))
        CTkToolTip(self.progressbar, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message="This is the training progress bar." +
                   " It will turn to full orange when the training was completed")
        
        self.progressbar.configure(mode="determinate", progress_color="orange")
        self.progressbar.set(0)
        
        
    # Show parameter explanation
    def __parameter_explanation(self):
        
        # Get link to the image
        image = Path(__file__).parents[1]
        button_image = ctk.CTkImage(Image.open( 
            Path(image, "images/parameter_explanation.png")),
            size=(800, 400))
            
        # Create new window 
        self.new_window = ctk.CTkToplevel(self, fg_color = "white")
        ctk.CTkLabel(master=self.new_window, image=button_image, text = "").pack()
            
    # Get number of epochs
    def get_nepoch(self, dummy):
        # Get number of hidden layers
        get_input_text = self.nepoch.get().strip()
        
        try:
            self.config["n_epochs"] = int(float(get_input_text))
            print("Number of epochs = ", self.config["n_epochs"])
            self.nepoch.configure(fg_color = '#d3ffc7')
        except:
            try:
                eval(get_input_text)
                self.config["n_epochs"] = get_input_text
                print("Number of epochs = ", self.config["n_epochs"])
                self.nepoch.configure(fg_color = '#d3ffc7')
            except:
                self.nepoch.configure(fg_color = '#ffc7c7')
                
    # Get learning_rate
    def get_learning_rate(self, dummy):
        
        get_input_text = self.learning_rate.get().strip()
        
        try:
            self.config["learning_rate"] = float(get_input_text)
            print("learning rate = ", self.config["learning_rate"])
            self.learning_rate.configure(fg_color = '#d3ffc7')
        except:
            try:
                eval(get_input_text)
                self.config["learning_rate"] = get_input_text
                print("learning rate = ", self.config["learning_rate"])
                self.learning_rate.configure(fg_color = '#d3ffc7')
            except:
                self.learning_rate.configure(fg_color = '#ffc7c7')
                

    # Get warm up length
    def get_warmup_length(self, dummy):

        get_input_text = self.warmup_length.get().strip()
        
        try:
            self.config["warmup_length"] = int(float(get_input_text))
            print("Warm up length = ", self.config["warmup_length"])
            self.warmup_length.configure(fg_color = '#d3ffc7')
        except:
            try:
                eval(get_input_text)
                self.config["warmup_length"] = get_input_text
                print("learning rate = ", self.config["warmup_length"])
                self.warmup_length.configure(fg_color = '#d3ffc7')
            except:
                self.warmup_length.configure(fg_color = '#ffc7c7')
                
                
    # Get learning_rate
    def get_sequence_length(self, dummy):
        
        get_input_text = self.sequence_length.get().strip()
        
        try:
            self.config["sequence_length"] = int(float(get_input_text))
            print("sequence length = ", self.config["sequence_length"])
            self.sequence_length.configure(fg_color = '#d3ffc7')
        except:
            try:
                eval(get_input_text)
                self.config["sequence_length"] = get_input_text
                print("learning rate = ", self.config["sequence_length"])
                self.sequence_length.configure(fg_color = '#d3ffc7')
            except:
                self.sequence_length.configure(fg_color = '#ffc7c7')
                

    # Get learning_rate
    def get_batch_size(self, dummy):
        
        get_input_text = self.batch_size.get().strip()
        
        try:
            self.config["batch_size"] = int(float(get_input_text))
            print("sequence length = ", self.config["batch_size"])
            self.batch_size.configure(fg_color = '#d3ffc7')
        except:
            try:
                eval(get_input_text)
                print("batch size = ", self.config["batch_size"])
                self.batch_size.configure(fg_color = '#d3ffc7')
            except:
                self.batch_size.configure(fg_color = '#ffc7c7')
                

    # Get learning_rate
    def get_patience_length(self, dummy):
        
        get_input_text = self.patience.get().strip()
        
        try:
            self.config["patience"] = int(float(get_input_text))
            print("sequence length = ", self.config["patience"])
            self.patience.configure(fg_color = '#d3ffc7')
        except:
            try:
                eval(get_input_text)
                self.config["patience"] = get_input_text
                print("patience length = ", self.config["patience"])
                self.patience.configure(fg_color = '#d3ffc7')
            except:
                self.patience.configure(fg_color = '#ffc7c7')
                
    # Get number of lstm layers
    def loss_function(self, method: str):
        loss_fn = {'Root Mean Square Error': "RMSE",
                   'Mean Absolute Error': "MAE",
                   'Mean Squared Error': "MSE",
                   '1 - Nash Sutcliffe': "NSE_complement"} 
                              
        self.config["loss_function"] = loss_fn[method]
        print(self.config["loss_function"])

    # Get number of lstm layers
    def get_optim_method(self, method: str):
        self.config["optim_method"] = method
        print(self.config["optim_method"])
        
    def out_dir_event(self):
        output_directory = tk.filedialog.askdirectory()
        self.config["output_directory"] = [output_directory]
        print("Output dir = ", self.config["output_directory"])
        
        # update text of the selected directory
        if len(output_directory) > 0:
            self.out_dir.configure(text = "..." + output_directory[-30:])
        else:
            self.out_dir.configure(text = "Select directory")
    
    def load_state_dict(self):
        # Select model state dict file
        state_dict_file = ctk.filedialog.askopenfilename(title="Select model state dict file", 
                                              filetypes=(('pt files', '*.pt'),
                                                         ('All files', '*.*')))
        # update text load model
        if state_dict_file[-3:] == ".pt":
            self.load_model.configure(text = "..." + state_dict_file[-30:])
            self.globalData['init_state_dicts'] = True
            self.globalData['init_state_dicts_file'] = state_dict_file
        else:
            self.load_model.configure(text = "Load model (optional)")
            self.globalData['init_state_dicts'] = False

    # Message box for error in input with tune search space 
    def message_box_tune(self, data_type):
        tk.messagebox.showinfo(
            title="Error",
            message="Input should be " + data_type + " or Tune Search Space API command: " +
            "https://docs.ray.io/en/latest/tune/api/search_space.html")

     
    def run_train_test(self):
        # Set progress to zero
        self.progressbar.set(0)

        self.run.configure(fg_color='gray')
        self.run.configure(state="disabled")
        
        tk.messagebox.showinfo(title="Message box", 
                               message="Trainning will start after closing this box")
        
        
        try:
            
            # Read data
            data = read_scale_data(self.config)
            
            # Convert config to search space
            search_space = config_to_search_space(
                self.config,
                data['x_train_scale'],
                data['y_train_scale'],
                data['x_valid_scale'],
                data['y_valid_scale']
                )
            
            if self.globalData['init_state_dicts'] and search_space["is_manual_optim"]:
                # Create the model
                model = create_model(self.config)
                
                model.load_state_dict(
                    torch.load(
                        self.globalData['init_state_dicts_file']
                        )
                    )
                
                # Train with train dataset
                trainer = Trainer(self.config, model)
                model = trainer.train(
                    data['x_train_scale'], 
                    data['y_train_scale'],
                    data['x_valid_scale'], 
                    data['y_valid_scale'])
        
                # Save train loss per epoch and best train loss
                data["loss_epoch"] = trainer.loss_epoch
                data["best_train_loss"] = trainer.best_train_loss
                best_config = self.config

            else:
                model, data, best_config = run_config(self.config)

            # Evaluate the model and transform back to normal scale
            data['y_train_simulated'] = data["y_scaler"].inverse(
                model.evaluate(data["x_train_scale"])
                )
            
            data['y_valid_simulated'] = data["y_scaler"].inverse(
                model.evaluate(data["x_valid_scale"])
                )
            
            data['y_test_simulated'] = data["y_scaler"].inverse(
                model.evaluate(data["x_test_scale"])
                )

            # Update global data
            self.globalData.update(data)
            self.globalData["model"] = model
            self.globalData["best_config"] = best_config
            self.globalData["config"] = self.config
            
 
            tk.messagebox.showinfo(title="Message box",
                                   message="Finished training/testing")
        except:
            tk.messagebox.showinfo(title="Error",
                                   message="Cannot train the model")

        self.progressbar.set(1.0)        
        self.run.configure(state="normal")
        self.run.configure(fg_color=['#3a7ebf', '#1f538d'])

    
        
        
        



from torch import nn
import torch
from hydroecolstm.model.linears import Linears


# LSTM + Linears

class MultiLinear(nn.Module):
    def __init__(self, input_size_1, input_size_2, output_size, activation_function):
        super(MultiLinear, self).__init__()
        self.Linear_1 = nn.Linear(input_size_1, output_size)
        self.Linear_2 = nn.Linear(input_size_2, output_size)
        self.activation_function = activation_function
        
    def forward(self, x_1, x_2):
        output = self.activation_function(self.Linear_1(x_1) + self.Linear_2(x_2))
        return output

class Ea_Lstm_Linears(nn.Module):
    def __init__(self, config):
        
        super(Ea_Lstm_Linears, self).__init__()
                
        self.static_size = len(config["input_static_features"])
        self.dynamic_size = len(config["input_dynamic_features"])
        self.num_layers = config["num_layers"]
        self.hidden_size = config["hidden_size"]
        self.output_size = len(config["target_features"])
        self.linears_num_layers = config["Regression"]["num_layers"]
        self.linears_activation_function = config["Regression"]["activation_function"]
        self.linears_num_neurons = self.find_num_neurons(config=config) 

        # Model structure
        self.i = nn.Sequential(nn.Linear(self.static_size, self.hidden_size), nn.Sigmoid())  
        self.f = MultiLinear(self.dynamic_size, self.hidden_size, self.hidden_size, nn.Sigmoid())
        self.g = MultiLinear(self.dynamic_size, self.hidden_size, self.hidden_size, nn.Tanh())
        self.o = MultiLinear(self.dynamic_size, self.hidden_size, self.hidden_size, nn.Sigmoid())
        self.linear = Linears(num_layers=self.linears_num_layers, 
                              activation_function=self.linears_activation_function,
                              num_neurons=self.linears_num_neurons)
        
    # TODO: This forward function takes too much times, need to improve
    def forward(self, x):
        # Initial hidden, cell states
        c_t = torch.randn(self.hidden_size).unsqueeze(0)
        h_t = torch.randn(self.hidden_size).unsqueeze(0)

        # Forward run
        output = {}
        for key in x.keys(): 
            output[key] = torch.zeros(size=(x[key].shape[0],self.output_size))
            for i in range(x[key].shape[0]):
                i_t = self.i(x[key][i:i+1,self.dynamic_size:])
                f_t = self.f(x[key][i:i+1,:self.dynamic_size], h_t)
                g_t = self.g(x[key][i:i+1,:self.dynamic_size], h_t)             
                o_t = self.o(x[key][i:i+1,:self.dynamic_size], h_t)
                
                c_t = f_t*c_t + i_t*g_t
                h_t = o_t*torch.tanh(c_t)
                
                output[key][i,:] = self.linear(h_t)

        return output
    
    # Find number of neuron in each linear layers, including the input layer
    def find_num_neurons(self, config) -> int:
        # First number of neurons from the input layers ()
        num_neurons = [self.hidden_size]

        if "Regression" in config:
            if len(config["Regression"]["num_neurons"]) > 1:
                for i in range(len(config["Regression"]["num_neurons"])-1):
                    num_neurons.append(config["Regression"]["num_neurons"][i])
        num_neurons.append(self.output_size)

        return num_neurons


import matplotlib.pyplot as plt

def plot(data: dict, object_id:str, target_feature:str):  
    # Get index of the target feature
    index = data["y_column_name"].index(target_feature)
    
    # Extract obeserved and target features from data
    y_observed_valid = data["y_valid"][object_id][:, index].detach().numpy()
    y_observed_test = data["y_test"][object_id][:, index].detach().numpy()
    
    
    y_valid_simulated = data['y_valid_simulated'][object_id][:, index].detach().numpy()
    y_test_simulated = data['y_test_simulated'][object_id][:, index].detach().numpy()
    
    time_valid = data["time_valid"][object_id]
    time_test = data["time_test"][object_id]
    
    
    # Now plot simulated and observed

    plt.plot(time_valid[5:], y_observed_valid[5:], color = 'red', label = "Observed (valid period)", alpha=0.1, linewidth=1)
    plt.plot(time_test[5:], y_observed_test[5:], color = 'red', label = "Observed (test period)", alpha=0.9, linewidth=1)   
    plt.plot(time_valid[5:], y_valid_simulated[5:], color = 'blue', label = "Simulated (valid period)", alpha=0.1, linewidth=1)
    plt.plot(time_test[5:], y_test_simulated[5:], color = 'blue', label = "Simulated (test period)", alpha=0.9, linewidth=1)
    plt.title(label=f"Catchment = {object_id}")
    plt.ylabel(target_feature)
    plt.legend()

    return plt

def plot_train_valid_loss(loss):
    #loss = data["loss"]
    plt.rcParams.update({'font.size': 8})
    
    epoch = loss.query('best_model == True').iloc[-1,:].epoch
    plt.plot(loss["epoch"], loss["train_loss"], label="Training loss")
    plt.plot(loss["epoch"], loss["validation_loss"], label="Validation loss")
    plt.axvline(x=epoch, label="Best model at epoch " + str(epoch), 
                color = "grey", linestyle='dashed')
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.legend()
    
    return plt



from .ea_lstm import Ea_Lstm_Linears
from .lstm_linears import Lstm_Linears
from .linears import Linears
from .create_model import create_model

__all__= ["Ea_Lstm_Linears", "Lstm_Linears", "Linears", "create_model"]





from torch import nn
from typing import List

class Linears(nn.Module):
    '''
    Multi-linear layers with different activation functions

    Parameters
    ----------
    num_layers : int
        Number of linear layers.
    activation_function: List[str]
        Name of  activation function for each linear layer, available function 
        names are "ReLu", "Sigmoid", "Tanh", "Softplus", "Identity"
    num_neurons: List[int]
        Number of neurons in each layers, including from the input size to the
        the multi-linear layers. Therefore, the len(num_neurons) = num_layers + 1

    Attributes
    ----------
    model : nn.Module
        The multi-linear layers
        
    Examples
    --------
    >>> import torch
    >>>
    >>> act_function_names = ["Identity", "Identity", "Identity"]
    >>> num_layers = 3
    >>> num_neurons = [2,3,4,3]
    >>>
    >>> x = torch.rand(10, num_neurons[0])
    >>> model = Linears(num_layers, act_function_names, num_neurons)
    >>>
    >>> # Forward run
    >>> model(x)
    '''
        
    def __init__(self, num_layers: int, activation_function: List[str], 
                 num_neurons: List[int]):
        super(Linears, self).__init__()
        
        # Activation functions
        activation_functions = {"ReLu": nn.ReLU(), "Sigmoid": nn.Sigmoid(), 
                                "Tanh": nn.Tanh(), "Softplus": nn.Softplus(), 
                                "Identity": nn.Identity()}
        
        # Create list to store different linear layers
        layers = []
        
        # Create layers of user-defined network
        for i in range(num_layers):
            layers.append(nn.Linear(num_neurons[i], num_neurons[i+1]))
            layers.append(activation_functions[activation_function[0]])
        
        # Combined all layers together using sequential
        self.model = nn.Sequential(*layers)
        
    # Forward run
    def forward(self, x):
        # model output
        output = self.model(x)
        
        # Return output
        return output
    
    


import torch

# scaler
class Scaler:
    def fit(self, x=None, method=None):
        # concatenat all object_id
        for i, object_id in zip(range(len(x)), x):
            if i == 0:
                cat_x = x[object_id]
            else:
                cat_x = torch.cat((cat_x, x[object_id]),0)
        
        # Get either min max or mean and standard deviation 
        self.mins = _column_mins(cat_x)
        self.maxs = _column_maxs(cat_x)
        
        # Get means and standar deviation
        self.means = _column_means(cat_x)
        self.stds = _column_stds(cat_x)
        
        scaler_a = []
        scaler_b = []
        
        for i, method_name in zip(range(len(method)), method):
            if method_name == "MinMaxScaler":
                scaler_a.append(self.mins[i])
                scaler_b.append(self.maxs[i] - self.mins[i])
            elif method_name=="Z-score":
                scaler_a.append(self.means[i])
                scaler_b.append(self.stds[i])
            elif method_name=="None":
                scaler_a.append(0.0)
                scaler_b.append(1.0)
            else:
                print("Error: unknown scaler")
                SystemExit("Program stop, please change scaler")
        
        scaler_ab = torch.cat((torch.tensor(scaler_a, dtype=torch.float32),
                               torch.tensor(scaler_b, dtype=torch.float32)), 0)
        
        self.scaler_parameter = torch.reshape(scaler_ab, 
                                              (2,len(scaler_a)))
 
    def transform(self, x:dict[str:torch.tensor]=None) -> list: 
        x_scale = {}
        for object_id in x:
            x_scale[object_id] =  torch.div(torch.sub(x[object_id], 
                                                      self.scaler_parameter[0,:]), 
                                            self.scaler_parameter[1,:])               
        return x_scale

    def inverse(self, x:list=None) -> list:        
        x_inverse = {}
        for object_id in x:
            x_inverse[object_id] =  torch.add(self.scaler_parameter[0,:],
                                              x[object_id]*self.scaler_parameter[1,:])
        return x_inverse

def _column_mins(input_tensor: torch.tensor=None):
    mask = ~torch.isnan(input_tensor) 
    colMins = [torch.min(torch.masked_select(input_tensor[:,i], mask[:,i]))
               for i in range(input_tensor.shape[1])]
    return torch.tensor(colMins, dtype=torch.float32)

def _column_maxs(input_tensor: torch.tensor=None):
    mask = ~torch.isnan(input_tensor) 
    colMaxs = [torch.max(torch.masked_select(input_tensor[:,i], mask[:,i]))
               for i in range(input_tensor.shape[1])]
    return torch.tensor(colMaxs, dtype=torch.float32)

def _column_means(input_tensor: torch.tensor=None):
    mask = ~torch.isnan(input_tensor) 
    colMeans = [torch.mean(torch.masked_select(input_tensor[:,i], mask[:,i]))
                for i in range(input_tensor.shape[1])]   
    return torch.tensor(colMeans, dtype=torch.float32)

def _column_stds(input_tensor: torch.tensor=None):
    mask = ~torch.isnan(input_tensor) 
    col_means = _column_means(input_tensor)
    
    col_stds = []
    
    for i in range(input_tensor.shape[1]):
        column = torch.masked_select(input_tensor[:,i], mask[:,i])
        col_stds.append((sum((column - col_means[i])**2)/len(column))**0.5)
    return torch.tensor(col_stds, dtype=torch.float32)

# Get scaler name in list
def get_scaler_name(config):
    
    if "input_static_features" not in config.keys():
        no_static_features = 0
    else:
        no_static_features = len(config["input_static_features"])
        
        
    # Get name of scaler for dynamic input
    scaler_name_input = config["scaler_input_dynamic_features"]*\
        len(config["input_dynamic_features"])
    
    # replicate this n times
    if no_static_features > 0 and\
        "scaler_input_static_features" in config:
            for name in config["scaler_input_static_features"]*no_static_features:
                scaler_name_input.append(name)
        
    # scaler name target
    scaler_name_target = config["scaler_target_features"]*len(config["target_features"])
    
    return scaler_name_input, scaler_name_target





from .read_config import read_config
from .read_data import read_train_valid_test_data, read_forecast_data, read_scale_data
from .custom_dataset import CustomDataset
from .scaler import Scaler

__all__= ["read_config", "read_train_valid_test_data", "read_forecast_data", 
          "read_scale_data", "CustomDataset", "Scaler"]




import pandas as pd
import numpy as np
import torch
from hydroecolstm.data.scaler import Scaler, get_scaler_name

# Read time series data into pandas data frame
def read_train_valid_test_data(config:dict=None) -> dict:
    
    # Read input data  
    dynamic_data = pd.read_csv(config['dynamic_data_file'][0], 
                               delimiter=",", header=0) 
    dynamic_data["time"] = pd.to_datetime(dynamic_data["time"], 
                                          format = "%Y-%m-%d %H:%M")
    
    # The column names must contains the following names
    require_columns = ["object_id","time"]
    require_columns.extend(config['input_dynamic_features'])
    require_columns.extend(config['target_features'])    
    
    # Check if data header contains required names        
    for name in require_columns: 
        if name not in dynamic_data.columns:
            raise NameError(f"Error: missing column '{name}' in dynamic data file \n")
            
    # Subset of dynamic_data - only use the required columns, rows
    dynamic_data = dynamic_data[require_columns]
    dynamic_data.set_index("object_id", inplace=True)
    dynamic_data = dynamic_data.loc[config["object_id"]]
    
    train_data = dynamic_data[(dynamic_data["time"] >= config["train_period"][0]) &
                              (dynamic_data["time"] <= config["train_period"][1])]

    # Colum name of the ouput tensor
    x_column_name = config['input_dynamic_features'].copy()
    y_column_name = config['target_features'].copy()
    
    # split to train data by object id    
    x_train = _split_by_object_id(train_data[x_column_name], config["object_id"])
    y_train = _split_by_object_id(train_data[y_column_name], config["object_id"])
    time_train = _time_by_object_id(train_data, config["object_id"])
 
    valid_data = dynamic_data[(dynamic_data["time"] >= config["valid_period"][0]) &
                             (dynamic_data["time"] <= config["valid_period"][1])]
    x_valid = _split_by_object_id(valid_data[x_column_name], config["object_id"])
    y_valid = _split_by_object_id(valid_data[y_column_name], config["object_id"])
    time_valid = _time_by_object_id(valid_data, config["object_id"])
    

    test_data = dynamic_data[(dynamic_data["time"] >= config["test_period"][0]) &
                             (dynamic_data["time"] <= config["test_period"][1])]
    x_test = _split_by_object_id(test_data[x_column_name], config["object_id"])
    y_test = _split_by_object_id(test_data[y_column_name], config["object_id"])
    time_test = _time_by_object_id(test_data, config["object_id"])
    
    # Read static input data file    
    if 'input_static_features' in config:
        if len(config['input_static_features']) > 0:
            static_data = pd.read_csv(config['static_data_file'][0], delimiter=",", 
                                      header=0)
            # The column names must contains the following names
            require_columns = ["object_id"]
            require_columns.extend(config['input_static_features'])    
            
            # Check if data header contains required names
            for name in require_columns: 
                if name not in static_data.columns:
                    raise NameError(f"Error: missing column '{name}' in static data\n")
            
            # Subset of dynamic_data - only use the required columns and rows
            static_data = static_data[require_columns]
            static_data.set_index("object_id", inplace=True)
            static_data = torch.tensor(static_data.loc[config["object_id"]].values,
                                       dtype=torch.float32)
            
            # Update columne name
            x_column_name.extend(config['input_static_features'])
            
    else:
        static_data = None
        
    # add static data to x_train and y_train
    if static_data is not None:
        for i, object_id in zip(range(len(x_train)), x_train):
            rep_static_data = static_data[i,].repeat(x_train[object_id].shape[0],1)
            x_train[object_id] = torch.cat((x_train[object_id], rep_static_data), 1)

            rep_static_data = static_data[i,].repeat(x_valid[object_id].shape[0],1)
            x_valid[object_id] = torch.cat((x_valid[object_id], rep_static_data), 1)
            
            rep_static_data = static_data[i,].repeat(x_test[object_id].shape[0],1)
            x_test[object_id] = torch.cat((x_test[object_id], rep_static_data), 1)

    return {"x_train":x_train, "y_train": y_train, "time_train" : time_train,
            "x_valid":x_valid, "y_valid": y_valid, "time_valid" : time_valid, 
            "x_test":x_test, "y_test": y_test, "time_test": time_test,
            "x_column_name": x_column_name, "y_column_name": y_column_name}

# -----------------------------------------------------------------------------
def read_forecast_data(config:dict=None) -> dict:
    
    # Read input data & check if users use the same file for train_test and forecast
    if config['dynamic_data_file_forecast'][0] == "dynamic_data_file": 
        dynamic_data = pd.read_csv(config['dynamic_data_file'][0], 
                                   delimiter=",", header=0)
    else:
        dynamic_data = pd.read_csv(config['dynamic_data_file_forecast'][0], 
                                   delimiter=",", header=0)
    
    # Convert date time to pandas date time
    dynamic_data["time"] = pd.to_datetime(dynamic_data["time"], 
                                          format = "%Y-%m-%d %H:%M")
   
    # The column names must contains the following names
    require_columns = ["object_id","time"]
    require_columns.extend(config['input_dynamic_features'])
    
    # Check if data header contains required names        
    for name in require_columns: 
        if name not in dynamic_data.columns:
            raise NameError(f"Error: missing column '{name}' in forecast data\n")

    # Add NaN to the dynamic data target features (if they are in the data) 
    for target_feature in config["target_features"]:
        if target_feature not in dynamic_data.columns:
            dynamic_data[target_feature] = np.nan
    require_columns.extend(config["target_features"])
    
    # Subset of dynamic_data - only selected columns
    dynamic_data = dynamic_data[require_columns]

    # Subset of dynamic data - only selected object_id
    dynamic_data.set_index("object_id", inplace=True)
    dynamic_data = dynamic_data.loc[config["object_id_forecast"]]
    
    # Subset of dynamic data - only forecast period
    forecast_data = dynamic_data[(dynamic_data["time"] >= config["forecast_period"][0]) &
                              (dynamic_data["time"] <= config["forecast_period"][1])]

    # Column name of the ouput tensor
    x_column_name = config['input_dynamic_features'].copy()
    y_column_name = config['target_features'].copy()
    
    # Split to x_forecast and y_forecast by object id
    x_forecast = _split_by_object_id(forecast_data[x_column_name], 
                                     config["object_id_forecast"])
    y_forecast = _split_by_object_id(forecast_data[y_column_name], 
                                     config["object_id_forecast"])
    time_forecast = _time_by_object_id(forecast_data, config["object_id_forecast"])
      
    # Read static input data file    
    if 'input_static_features' in config:
        if len(config['input_static_features']) > 0:
            if config['static_data_file_forecast'][0] == "static_data_file":
                static_data = pd.read_csv(config['static_data_file'][0], 
                                          delimiter=",", header=0)
            else:
                static_data = pd.read_csv(config['static_data_file_forecast'][0], 
                                          delimiter=",", header=0)
                
            # The column names must contains the following names
            require_columns = ["object_id"]
            require_columns.extend(config['input_static_features'])    
            
            # Check if data header contains required names
            for name in require_columns: 
                if name not in static_data.columns:
                    raise NameError(f"Error: missing column '{name}' in" + 
                                    " static data file \n")
            
            # Subset of dynamic_data - only use the required columns and rows
            static_data = static_data[require_columns]
            static_data.set_index("object_id", inplace=True)
            static_data = torch.tensor(static_data.loc[config["object_id_forecast"]].values,
                                       dtype=torch.float32)
            
            # Update columne name
            x_column_name.extend(config['input_static_features'])
            
    else:
        static_data = None

    # add static data to x_forecast and y_forecast
    if static_data is not None:
        for i, object_id in zip(range(len(x_forecast)), x_forecast):
            rep_static_data = static_data[i,].repeat(x_forecast[object_id].shape[0],1)
            x_forecast[object_id] = torch.cat((x_forecast[object_id], rep_static_data), 1)

    return {"x_forecast":x_forecast, "y_forecast": y_forecast, 
            "time_forecast": time_forecast, "x_column_name": x_column_name, 
            "y_column_name": y_column_name}

# split by object id
def _split_by_object_id(data, object_id):
    output = {}
    for objectid in object_id: 
        output[str(objectid)] = torch.tensor(data.loc[objectid].values,
                                             dtype=torch.float32)
    return output

# Get date time by object id
def _time_by_object_id(data, object_id):
    output = {}
    for objectid in object_id: 
        output[str(objectid)] = data.loc[objectid]["time"].values
    return output

# Read and scale data
def read_scale_data(config):
    data = read_train_valid_test_data(config)
    
    # Scale/transformer name for static, dynamic, and target features
    x_scaler_name, y_scaler_name = get_scaler_name(config)
    
    # Scaler/transformer
    data['x_scaler'], data['y_scaler'] = Scaler(), Scaler()
    data['x_scaler'].fit(x=data["x_train"], method=x_scaler_name)
    data['y_scaler'].fit(x=data["y_train"], method=y_scaler_name)
    
    # Scale/transform data input data
    data['x_train_scale'] = data['x_scaler'].transform(x=data["x_train"])
    data['x_valid_scale'] = data['x_scaler'].transform(x=data["x_valid"])
    data['x_test_scale'] = data['x_scaler'].transform(x=data["x_test"])
    
    # Scale/transform data target train data    
    data['y_train_scale'] = data['y_scaler'].transform(x=data["y_train"])
    data['y_valid_scale'] = data['y_scaler'].transform(x=data["y_valid"])  
    
    return data



import customtkinter as ctk
import torch
import matplotlib
import tkinter as tk
from matplotlib.figure import Figure
import numpy as np
import pandas as pd
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, 
                                               NavigationToolbar2Tk)
#import matplotlib.pyplot as plt 

# Plot loss
def plot_train_valid_loss(loss, plot_window, xlabel, ylabel, title, 
                          train_color, train_line_style, train_legend,
                          valid_color, valid_line_style, valid_legend,
                          best_model_legend):

    figure = Figure(figsize=(15, 4.5), dpi=100)
    figure_canvas = FigureCanvasTkAgg(figure, plot_window)
    toolbar = NavigationToolbar2Tk(figure_canvas, plot_window, 
                                   pack_toolbar=False)
    toolbar.update()
    toolbar.pack(side=tk.BOTTOM, fill=tk.X)
            
    axes = figure.add_subplot()
    
    # Plus 1 to best epoch number as the epoch index starts from 0
    epoch = np.where(loss.best_model == True)[0][0] + 1
    
    # Default plot setting
    
    if train_color == "": train_color = "coral"
    if train_line_style == "": train_line_style = "dashed"
    if train_legend == "": train_legend = "Training loss"
    if valid_color == "": valid_color = "blue"
    if valid_line_style == "": valid_line_style = "solid"
    if valid_legend == "": valid_legend = "Validation loss"
    if xlabel == "": xlabel = "Epoch"
    if ylabel == "": ylabel = "Loss"
    if title == "": title = 'Training and validation loss'
    if best_model_legend == "": best_model_legend = "Best model at epoch " + str(epoch)
    
    try:
        axes.plot(loss.index + 1, 
                  loss["train_loss"], 
                  label = train_legend, 
                  linestyle = train_line_style,
                  color = train_color)
    except:
        axes.plot(loss.index + 1, 
                  loss["train_loss"], 
                  label = train_legend, 
                  linestyle = "dashed",
                  color = "coral")
    try:    
        axes.plot(loss.index + 1, 
                  loss["valid_loss"], 
                  label=valid_legend, 
                  linestyle = valid_line_style,
                  color = valid_color)
    except:
        axes.plot(loss.index + 1, 
                  loss["valid_loss"], 
                  label=valid_legend, 
                  color = "blue")
    axes.axvline(x=epoch, 
                 label=best_model_legend,
                 color = "grey",
                 linestyle='dashed')
    
    # Set x label
    axes.set_xlabel(xlabel)
    axes.set_ylabel(ylabel)
    axes.set_title(title)
    axes.legend()
            
    figure_canvas.get_tk_widget().pack(side=tk.TOP, 
                                       fill=tk.BOTH, 
                                       expand=1)

# Check color
def check_color(entry, default):
    color  = entry.get().strip()
    if len(color) > 0:
        if color not in matplotlib.colors.cnames.keys():
            tk.messagebox.showinfo(title="Error", message= color +
                                   ": unknown color name, default color is used")
            color = default 
    else:
        color = default
    return color

# Check color
def check_marker(entry, default):
    marker  = entry.get().strip()
    if len(marker) > 0:
        if marker not in [".","o","s","^","v","+","x"]:
            tk.messagebox.showinfo(title="Error", message= marker +
                                   ": unknown marker name, default marker is used")
            marker = default 
    else:
        marker = default
    return marker

# Check color
def check_line_plot(entry):
    line_plot  = entry.get().strip()
    if len(line_plot) > 0:
        if line_plot in ["False","false","FALSE"]:
            line_plot = False
        else:
            line_plot = True
    else:
        line_plot = True
    return line_plot
    
# Check alpha
def check_alpha(entry, default):
    alpha =  entry.get().strip()
    
    if len(alpha) > 0:
        try:
            alpha = float(alpha)
            if alpha > 1.0: alpha = 1.0
            if alpha < 0.0: alpha = 0.0
        except:
            alpha = default
    else:
        alpha = default
    return alpha

# Check line style    
def check_linestyle(entry, default):
    linestyle =  entry.get().strip()
    
    if len(linestyle) > 0:
        if linestyle not in ["solid", "dashed", "dashdot", "dotted"]:
            linestyle = default
    else:
        linestyle = default
        
    return linestyle

# Check line style    
def check_size(entry, default):
    size =  entry.get().strip()
    
    if len(size) > 0:
        try:
            size = float(size)
            if size < 0.01: size = default
        except:
            size = default
    else:
        size = default
    return size

# Check line style    
def check_ylim(entry, default):
    ylim =  entry.get().strip()
    
    if len(ylim) > 0:
        try:
            ylim = float(ylim)
        except:
            ylim = False
    else:
        ylim = False
    return ylim



#-----------------------------------------------------------------------------#
#                                 plot timeseries                             #
#-----------------------------------------------------------------------------#
def plot_time_series(plot_window, data, key, idx, lineplot, 
                 color_obs, alpha_obs, size_obs, linestyle_obs,
                 marker_obs, label_obs, color_sim, alpha_sim, 
                 size_sim, linestyle_sim, label_sim, title, xlabel,
                 ylabel, ylim_up, forecast_period):
    
    # Create figure
    figure = Figure(figsize=(15, 4.5), dpi=100)
    figure_canvas = FigureCanvasTkAgg(figure, plot_window)
    toolbar = NavigationToolbar2Tk(figure_canvas, plot_window, 
                                   pack_toolbar=False)
    toolbar.update()
    toolbar.pack(side=tk.BOTTOM, fill=tk.X)
    axes = figure.add_subplot()
    
    if not forecast_period:
        # Key names of time and y variables in the input data
        time_name = ["time_train", "time_valid", "time_test"]
        period = ["Training", "Validation", "Test"]
        y_name = ["y_train", "y_valid", "y_test"]
        
        # Plot observed -----------------------------------------------------------
        for i in range(3):
            time = data[time_name[i]][key]
            y_obs = data[y_name[i]][key][:,idx].detach().numpy()
            
            if lineplot:
                if i < 2:
                    axes.plot(time, 
                              y_obs, 
                              linestyle=linestyle_obs,
                              color=color_obs, 
                              alpha=alpha_obs,
                              linewidth=size_obs)
                else:
                    axes.plot(time, 
                              y_obs, 
                              linestyle=linestyle_obs,
                              color=color_obs, 
                              alpha=alpha_obs, 
                              linewidth=size_obs, 
                              label=label_obs)                
            else:
                if i < 2:
                    axes.plot(time, 
                              y_obs, 
                              marker_obs,
                              color=color_obs, 
                              alpha=alpha_obs,
                              markersize=size_obs)
                else:
                    axes.plot(time, 
                              y_obs, 
                              marker_obs,
                              color=color_obs, 
                              alpha=alpha_obs,
                              markersize=size_obs,
                              label=label_obs)
                
                
        # Plot simulated ----------------------------------------------------------   
        for i in range(3):
            time = data[time_name[i]][key]
            y_sim = data[y_name[i]+"_simulated"][key][:,idx].detach().numpy()
    
            if i < 2:
                axes.plot(time, 
                          y_sim, 
                          linestyle=linestyle_sim,
                          color=color_sim, 
                          alpha=alpha_sim,
                          linewidth=size_sim)
            else:
                axes.plot(time, 
                          y_sim, 
                          linestyle=linestyle_sim,
                          color=color_sim, 
                          alpha=alpha_sim,
                          linewidth=size_sim,
                          label=label_sim)
            # Add vertical lines seperate train, valid, test
            axes.axvline(x=time[-1], color = "grey", linestyle='dashed')
            
            # Add text train, valid, test
            if i == 0: 
                ylable = np.nanmax(y_sim) + 0.6*np.absolute(np.nanmax(y_sim))
                
            axes.text(time[1], ylable, period[i])
    else:
        time = data["time_forecast"][key]
        y_obs = data["y_forecast"][key][:,idx].detach().numpy()
        y_sim = data["y_forecast_simulated"][key][:,idx].detach().numpy()
        
        # Plot observed
        if lineplot:
            axes.plot(time,
                      y_obs, 
                      linestyle=linestyle_obs,
                      color=color_obs, 
                      alpha=alpha_obs, 
                      linewidth=size_obs, 
                      label=label_obs)                
        else:
            axes.plot(time,
                      y_obs,
                      marker_obs,
                      color=color_obs,
                      alpha=alpha_obs,
                      markersize=size_obs,
                      label=label_obs)
            
        # Plot simulated
        axes.plot(time,
                  y_sim, 
                  linestyle=linestyle_sim,
                  color=color_sim, 
                  alpha=alpha_sim,
                  linewidth=size_sim,
                  label=label_sim)
        
    # Set x label
    axes.set_xlabel(xlabel)
    axes.set_ylabel(ylabel)
    axes.set_title(title)
    
    if type(ylim_up) is not bool:
        axes.set_ylim(top=ylim_up)

    axes.legend()
    
        
    figure_canvas.get_tk_widget().pack(side=tk.TOP, 
                                       fill=tk.BOTH, 
                                       expand=1)
       
           
def config_to_text(config):
    out_text = []
    for key in config.keys():   
        # Write list object in multiple lines             
        if type(config[key]) is list:
            out_text.append(key + ":\n")
            for element in config[key]:
                out_text.append("  - " + str(element) + "\n")
                
        elif type(config[key]) is dict:
            config_key = config[key]
            out_text.append(key + ":\n")
            
            for key in config_key.keys():
                if type(config_key[key]) is list:
                    out_text.append("  " + key + ":\n")
                    for element in config_key[key]:
                        out_text.append("    - " + str(element) + "\n")
                else:
                    out_text.append("  " + key +": " + str(config_key[key]) + "\n")
        else:
            try:
                # Convert time in config to YYYY-MM-DD HH:MM
                if (config[key].shape[0] == 2):
                    out_text.append(key +": \n")
                    if key == "train_period":
                        out_text.append("  - " + str(config["train_period"][0])[:16] + "\n")
                        out_text.append("  - " + str(config["train_period"][1])[:16] + "\n")
                    elif key == "valid_period":
                        out_text.append("  - " + str(config["valid_period"][0])[:16] + "\n")
                        out_text.append("  - " + str(config["valid_period"][1])[:16] + "\n")
                    else:
                        out_text.append("  - " + str(config["test_period"][0])[:16] + "\n")
                        out_text.append("  - " + str(config["test_period"][1])[:16] + "\n")                           
            except:
                # Non list object writte in 1 line
                out_text.append(key +": " + str(config[key]) + "\n")
                #out_text.append("\n")
                
    return out_text
    
def sort_key(config):
    config_sort = {}
        
    if "dynamic_data_file" in config.keys():
        config_sort["dynamic_data_file"] = config["dynamic_data_file"]

    if "static_data_file" in config.keys():
        config_sort["static_data_file"] = config["static_data_file"]

    if "output_directory" in config.keys():
        config_sort["output_directory"] = config["output_directory"] 
        
    if "input_static_features" in config.keys():
        config_sort["input_static_features"] = config["input_static_features"] 

    if "input_dynamic_features" in config.keys():
        config_sort["input_dynamic_features"] = config["input_dynamic_features"]        

    if "target_features" in config.keys():
        config_sort["target_features"] = config["target_features"] 

    if "object_id" in config.keys():
        config_sort["object_id"] = config["object_id"]        

    if "train_period" in config.keys():
        config_sort["train_period"] = config["train_period"]

    if "valid_period" in config.keys():
        config_sort["valid_period"] = config["valid_period"]
        
    if "test_period" in config.keys():
        config_sort["test_period"] = config["test_period"]
        
    if "model_class" in config.keys():
        config_sort["model_class"] = config["model_class"]

    if "Regression" in config.keys():
        config_sort["Regression"] = config["Regression"]
        
    if "Regression" in config.keys():
        config_sort["Regression"] = config["Regression"]

    if "scaler_input_dynamic_features" in config.keys():
        config_sort["scaler_input_dynamic_features"] = config["scaler_input_dynamic_features"]        

    if "scaler_input_static_features" in config.keys():
        config_sort["scaler_input_static_features"] = config["scaler_input_static_features"] 

    if "scaler_target_features" in config.keys():
        config_sort["scaler_target_features"] = config["scaler_target_features"]        

    if "hidden_size" in config.keys():
        config_sort["hidden_size"] = config["hidden_size"] 

    if "num_layers" in config.keys():
        config_sort["num_layers"] = config["num_layers"]        
        
    if "n_epochs" in config.keys():
        config_sort["n_epochs"] = config["n_epochs"] 

    if "learning_rate" in config.keys():
        config_sort["learning_rate"] = config["learning_rate"]        

    if "dropout" in config.keys():
        config_sort["dropout"] = config["dropout"] 

    if "warmup_length" in config.keys():
        config_sort["warmup_length"] = config["warmup_length"]        

    if "optim_method" in config.keys():
        config_sort["optim_method"] = config["optim_method"] 

    if "loss_function" in config.keys():
        config_sort["loss_function"] = config["loss_function"]        

    if "sequence_length" in config.keys():
        config_sort["sequence_length"] = config["sequence_length"]

    if "batch_size" in config.keys():
        config_sort["batch_size"] = config["batch_size"]
        
    if "patience" in config.keys():
        config_sort["patience"] = config["patience"]

    if "static_data_file_forecast" in config.keys():
        config_sort["static_data_file_forecast"] = config["static_data_file_forecast"]        

    if "dynamic_data_file_forecast" in config.keys():
        config_sort["dynamic_data_file_forecast"] = config["dynamic_data_file_forecast"]        
        
    if "forecast_period" in config.keys():
        config_sort["forecast_period"] = config["forecast_period"] 

    if "object_id_forecast" in config.keys():
        config_sort["object_id_forecast"] = config["object_id_forecast"]
            
    return config_sort

def write_yml_file(config, out_file):
    # Convert config to text
    output_text = config_to_text(config=sort_key(config))
        
    # Write config to config file
    with open(out_file, "w") as config_file:
        for line in output_text:
            config_file.write(line)
            
class ToplevelWindow(ctk.CTkToplevel):
    def __init__(self, window_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.geometry("600x400")
        self.title(window_name)
        
def combine_simulated(data, target_features):
    
    # combine train, valid, test results
    for i, key in enumerate(data["y_train_simulated"].keys()):
        
        # simulated target variables during train, valid, test periods
        sim = torch.cat((data["y_train_simulated"][key],
                         data["y_valid_simulated"][key],
                         data["y_test_simulated"][key]))
        
        obs = torch.cat((data["y_train"][key],
                         data["y_valid"][key],
                         data["y_test"][key]))
        
        # date
        date = np.concatenate((data["time_train"][key],
                               data["time_valid"][key],
                               data["time_test"][key]), axis=0)
        
        # Object id
        train_length = data["y_train_simulated"][key].shape[0]
        valid_length = data["y_valid_simulated"][key].shape[0]
        test_length = data["y_test_simulated"][key].shape[0]
        
        object_id = np.repeat(key, train_length+valid_length+test_length, axis=None)
        flag = np.concatenate((np.repeat("train_period", train_length, axis=None), 
                               np.repeat("valid_period", valid_length, axis=None),
                               np.repeat("test_period", test_length, axis=None)), axis=0) 
        
        # Now combine
        if i == 0:
            cat_sim = sim
            cat_obs = obs
            cat_date = date
            cat_object_id = object_id
            cat_flag = flag
        else:
            cat_sim = torch.cat((cat_sim, sim))
            cat_obs = torch.cat((cat_obs, obs))
            cat_date = np.concatenate((cat_date, date), axis=0)
            cat_object_id = np.concatenate((cat_object_id, object_id), axis=0)
            cat_flag = np.concatenate((cat_flag, flag), axis=0)
    
    # combine simulated and observed
    cat_sim_obs = torch.cat((cat_sim, cat_obs), axis=1)
    
    # convert to data frame
    output = pd.DataFrame(data=cat_sim_obs.numpy())
    
    # rename data frame
    colnames = np.concatenate((["simulated_" + name for name in target_features],
                                      ["observed_" + name for name in target_features]), 
                                     axis=0)
    output = output.set_axis(colnames, axis='columns') 
    
    # insert 
    output.insert(0, "flag", cat_flag.ravel(), True)
    output.insert(0, "time", cat_date.ravel(), True)
    output.insert(0, "object_id", cat_object_id.ravel(), True)
    
    return output

def combine_forecast(data, target_features):
    
    # combine train, valid, test results
    for i, key in enumerate(data["y_forecast"].keys()):
        
        # simulated target variables during train, valid, test periods
        sim = data["y_forecast_simulated"][key]
        date = data["time_forecast"][key]
        object_id = np.repeat(key, data["y_forecast_simulated"][key].shape[0], 
                              axis=None)
        
        # Now combine
        if i == 0:
            cat_sim = sim
            cat_date = date
            cat_object_id = object_id
        else:
            cat_sim = torch.cat((cat_sim, sim))
            cat_date = np.concatenate((cat_date, date), axis=0)
            cat_object_id = np.concatenate((cat_object_id, object_id), axis=0)
    
    # convert to data frame
    output = pd.DataFrame(data=cat_sim.numpy())
    output = output.set_axis(["forecast_" + name for name in target_features], 
                             axis='columns') 
    
    # insert 
    output.insert(0, "time", cat_date.ravel(), True)
    output.insert(0, "object_id", cat_object_id.ravel(), True)
    
    return output


import torch
from torch.utils.data import Dataset


class CustomDataset(Dataset):
    def __init__(self, 
                 X:dict[str:torch.Tensor], 
                 Y:dict[str:torch.Tensor], 
                 sequence_length:int, 
                 overlap:int=0):
        
        self.X = None
        self.Y = None
        self.overlap = overlap
        self.sequence_length = sequence_length   
        
        # Combine 2D time series of all basins (keys)
        for key in X.keys():
            Y[key][:self.overlap,:] = float("nan")
            if self.X is None:
                self.X = X[key]
                self.Y = Y[key]
            else:
                self.X = torch.cat((self.X, X[key]), dim = 0)
                self.Y = torch.cat((self.Y, Y[key]), dim = 0)
                    
    def __getitem__(self, index):
        # Get starting and ending index of each mini-batch
        istart = index*(self.sequence_length - self.overlap)
        iend = istart + self.sequence_length
            
        # Mini batch data
        if iend <= len(self.X):
            _x = self.X[istart:iend,:]
            _y = self.Y[istart:iend,:]
        else:
            _x = self.X[-self.sequence_length:,:]
            _y = self.Y[-self.sequence_length:,:]
        
        # Return mini batch consist of input, output
        return _x, _y
    
    def __len__(self):
        # The number of batches (round up)
        num_batches = int(round((len(self.X) - self.overlap)/
                          (self.sequence_length - self.overlap) + 0.49))                          
        return num_batches




import customtkinter as ctk
from CTkToolTip import CTkToolTip

class SidebarFrame(ctk.CTkFrame):
    def __init__(self, container=None, data_frame=None, network_frame=None,
                 train_test_frame=None, visual_frame=None, application_frame=None):
        super().__init__(container)
        
        # setup the grid layout manager
        self.data_frame = data_frame
        self.network_frame = network_frame
        self.train_test_frame = train_test_frame
        self.visual_frame = visual_frame
        self.application_frame = application_frame
        self.columnconfigure(0, weight=1)
        self.rowconfigure((0,1,2,3,4,5,6,7,8), weight=0)
        self.__create_widgets() 
    
    # create widgets for sidebar frame
    def __create_widgets(self):     
        self.label = ctk.CTkLabel(self, 
                                  text="HydroEcoLSTM", 
                                  font=ctk.CTkFont(size=20, weight="bold"))
        self.label.grid(row=0, column=0, padx=25, pady=20)
        CTkToolTip(self.label, delay=0.1, bg_color = 'orange', 
                   text_color = 'black', anchor = 'w', 
                   message= 'ALWAYS follow the numbering scheme (1 >> 2 >> 3' +
                   ' and so on) to create your project')

        # Data processing
        self.data_button = ctk.CTkButton(self, anchor='w', 
                                         command=self.data_button_event,
                                         text="1. Data Processing")
        self.data_button.grid(row=1, column=0, padx=0, pady=10)
        CTkToolTip(self.data_button, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click this button for preparing your ' +
                   'data input and target data\n')
        
        # Network definition
        self.net_button = ctk.CTkButton(self, anchor='w', 
                                        command=self.network_button_event,
                                        text="2. Network Design")
        self.net_button.grid(row=2, column=0, padx=0, pady=10)
        CTkToolTip(self.net_button, delay=0.1, bg_color = 'orange', 
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to design your neural network architect')
        
        # Training and testing the network
        self.train_button = ctk.CTkButton(self,
                                          command=self.traintest_button_event,
                                          anchor='w', 
                                          text="3. Training/Testing")
        self.train_button.grid(row=3, column=0, padx=0, pady=10)
        CTkToolTip(self.train_button, delay=0.1, bg_color = 'orange', 
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to set parameters for trainning and testing your network')

        self.visualize_button = ctk.CTkButton(self,
                                          command=self.visualize_button_event,
                                          anchor='w', 
                                          text="4. Visualize results")
        self.visualize_button.grid(row=4, column=0, padx=0, pady=10)
        CTkToolTip(self.visualize_button, delay=0.1, bg_color = 'orange', 
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to visualize the model outputs')

        self.application_button = ctk.CTkButton(self,
                                          command=self.application_button_event,
                                          anchor='w', 
                                          text="5. Application")
        self.application_button.grid(row=5, column=0, padx=0, pady=10)
        CTkToolTip(self.application_button, delay=0.1, bg_color = 'orange', 
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to make a forward run for forcast, or \n' +
                       'simulation in ungauged basins using the trained network')
        
        # Appearance
        self.appearance_option = ctk.CTkOptionMenu(self,
                                                   values=["Light", "Dark", "System"],
                                                   button_color='gray', fg_color = 'gray',
                                                   command=self.appearance_event) 
        self.appearance_option.place(relx=0.5, rely=0.90, anchor="n")
        CTkToolTip(self.appearance_option, delay=0.1, bg_color = 'orange', 
                   text_color = 'black', anchor='w', 
                   message='Please select your background themes here')
        
        self.scaling_optionemenu = ctk.CTkOptionMenu(self, 
                                                     values=["110%", "75%", "100%", "125%", "150%"],
                                                     button_color='gray', fg_color = 'gray',
                                                     command=self.change_scaling_event)
        self.scaling_optionemenu.place(relx=0.5, rely=0.82, anchor="n")
        CTkToolTip(self.scaling_optionemenu, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Please select the text size here \n')
        
        # Only activate the data_button when open GUI
        self.data_button.configure(fg_color=['#3a7ebf', '#1f538d'])
        self.net_button.configure(fg_color='gray')
        self.train_button.configure(fg_color='gray')
        self.visualize_button.configure(fg_color='gray')
        self.application_button.configure(fg_color='gray')
     
    def change_scaling_event(self, new_scaling: str):
        new_scaling_float = int(new_scaling.replace("%", "")) / 100
        ctk.set_widget_scaling(new_scaling_float)
        
    # button event in sidebar frame 
    def appearance_event(self, new_appearance: str):
        ctk.set_appearance_mode(new_appearance)
        
    def data_button_event(self):
        # turn of other frames
        self.network_frame.grid_forget()
        self.train_test_frame.grid_forget()
        self.visual_frame.grid_forget()
        self.application_frame.grid_forget()
        
        # turn on data frame
        self.data_frame.grid(row=0, column=1, rowspan=4, padx = 10, 
                               columnspan=2, pady=(20, 20), sticky="nsew")
        
        # change button color
        self.data_button.configure(fg_color=['#3a7ebf', '#1f538d'])
        self.net_button.configure(fg_color='gray')
        self.train_button.configure(fg_color='gray')
        self.visualize_button.configure(fg_color='gray')
        self.application_button.configure(fg_color='gray')
        
    def network_button_event(self):
        # turn of other frames
        self.data_frame.grid_forget()
        self.train_test_frame.grid_forget()
        self.visual_frame.grid_forget()
        self.application_frame.grid_forget()
        
        # turn on data frame
        self.network_frame.grid(row=0, column=1, rowspan=4, padx = 10, 
                               columnspan=2, pady=(20, 20), sticky="nsew")
        
        # change button color
        self.data_button.configure(fg_color='gray')
        self.net_button.configure(fg_color=['#3a7ebf', '#1f538d'])
        self.train_button.configure(fg_color='gray')
        self.visualize_button.configure(fg_color='gray')
        self.application_button.configure(fg_color='gray')
        
    def traintest_button_event(self):
        # turn of other frames
        self.data_frame.grid_forget()
        self.network_frame.grid_forget()
        self.visual_frame.grid_forget()
        self.application_frame.grid_forget()
        
        # turn on data frame
        self.train_test_frame.grid(row=0, column=1, rowspan=4, padx = 10, 
                               columnspan=2, pady=(20, 20), sticky="nsew")
        
        # change button color
        self.data_button.configure(fg_color='gray')
        self.train_button.configure(fg_color=['#3a7ebf', '#1f538d'])
        self.net_button.configure(fg_color='gray')
        self.visualize_button.configure(fg_color='gray')
        self.application_button.configure(fg_color='gray')
        
    def visualize_button_event(self):
        # turn of other frames
        self.data_frame.grid_forget()
        self.network_frame.grid_forget()
        self.train_test_frame.grid_forget()
        self.application_frame.grid_forget()
        
        # turn on data frame
        self.visual_frame.grid(row=0, column=1, rowspan=4, padx = 10, 
                               columnspan=2, pady=(20, 20), sticky="nsew")
        
        # change button color
        self.data_button.configure(fg_color='gray')
        self.train_button.configure(fg_color='gray')
        self.net_button.configure(fg_color='gray')
        self.visualize_button.configure(fg_color=['#3a7ebf', '#1f538d'])
        self.application_button.configure(fg_color='gray')
        
    def application_button_event(self):
        self.data_frame.grid_forget()
        self.network_frame.grid_forget()
        self.train_test_frame.grid_forget()
        self.visual_frame.grid_forget()
   
        # turn on data frame
        self.application_frame.grid(row=0, column=1, rowspan=4, padx = 10, 
                               columnspan=2, pady=(20, 20), sticky="nsew")
        
        # change button color
        self.data_button.configure(fg_color='gray')
        self.train_button.configure(fg_color='gray')
        self.net_button.configure(fg_color='gray')
        self.visualize_button.configure(fg_color='gray')
        self.application_button.configure(fg_color=['#3a7ebf', '#1f538d'])
                
        
        
        



from torch import nn
import torch
from hydroecolstm.model.linears import Linears

# LSTM + Linears
class Lstm_Linears(nn.Module):
    def __init__(self, config, **kwargs):
        
        super(Lstm_Linears, self).__init__()

        self.input_size = self.get_input_size(config)
        self.output_size = len(config["target_features"])
        self.hidden_size = config["hidden_size"]
        self.num_layers = config["num_layers"]
        self.dropout = config["dropout"]*min(1.0, self.num_layers - 1.0)
        self.linears_num_layers = config["Regression"]["num_layers"]
        self.linears_activation_function = config["Regression"]["activation_function"]
        self.linears_num_neurons = self.find_num_neurons(config=config) 
        
        # Standard LSTM from torch
        self.lstm = nn.LSTM(input_size=self.input_size, 
                            hidden_size=self.hidden_size, 
                            num_layers=self.num_layers,
                            dropout=self.dropout,
                            **kwargs)


        # Fully-connected layer connect hidden and output
        self.linear = Linears(num_layers=self.linears_num_layers, 
                              activation_function=self.linears_activation_function,
                              num_neurons=self.linears_num_neurons)
     
    # Forward mode run with torch.Tensor as usrual
    def forward(self, x:torch.Tensor) -> torch.Tensor:
        
        # get standard LSTM outputs
        y_predict, _ = self.lstm(x)
        
        # get final output 
        y_predict = self.linear(y_predict)  
        
        # return output
        return y_predict

    # In the evaluate mode run with Dict[str: tensor]
    def evaluate(self, x:dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        
        with torch.no_grad():
            y_predict = {}
            
            for key in x.keys():
                # get standard LSTM outputs
                y_predict[key], _ = self.lstm(x[key])
                
                # get final output 
                y_predict[key] = self.linear(y_predict[key])  
        
        # return output
        return y_predict

    
    # get input size
    def get_input_size(self, config) -> int:
        if "input_static_features" in config:
            input_size = (len(config["input_dynamic_features"]) + 
                          len(config["input_static_features"]))
        else:
            input_size = len(config["input_dynamic_features"])
        return input_size
    
    # Find number of neuron in each linear layers, including the input layer
    def find_num_neurons(self, config) -> int:
        # First number of neurons from the input layers ()
        num_neurons = [self.hidden_size]

        if "Regression" in config:
            if len(config["Regression"]["num_neurons"]) > 1:
                for i in range(len(config["Regression"]["num_neurons"])-1):
                    num_neurons.append(config["Regression"]["num_neurons"][i])
        num_neurons.append(self.output_size)

        return num_neurons



import customtkinter as ctk
import tkinter as tk
from CTkToolTip import CTkToolTip
from PIL import Image
from pathlib import Path
from ray import tune

class NetworkDesignFrame(ctk.CTkScrollableFrame):
    def __init__(self, container=None, config=None, globalData=None):
        super().__init__(container)
        
        # setup the grid layout manager
        self.config = config
        self.globalData = globalData
        self.columnconfigure((0), weight=1)
        self.__create_widgets()

        
    # create widgets for sidebar frame
    def __create_widgets(self): 
        
        # create tabs
        self.tabview = ctk.CTkTabview(master=self, width = 750, 
                                      border_width=1.5,
                                      fg_color = "transparent")
        self.tabview.pack(fill='both',expand=1)
        self.tabview.add("1. Model class")
        self.tabview.tab("1. Model class").grid_columnconfigure((0,1), weight=1)
        self.tabview.tab("1. Model class").rowconfigure((0,1,2,3,4,5,6,7), weight=1)
        self.tabview.add("2. Model head")
        self.tabview.tab("2. Model head").grid_columnconfigure((0), weight=1)
        
        self.ray_tooltip = ('Or input the tune Search Space API command, ' +
                            'e.g., tune.grid_search([20, 30]). Please see: ' + 
                            'https://docs.ray.io/en/latest/tune/api/search_space.html \n'+
                            'IMPORTANT: Box color interpretation: \n ' +
                            'wrong input = light rose  \n ' +
                            'correct input = light green color')
        
        ctk.CTkButton(master=self,  anchor='w', fg_color='gray',
                      text="Show model head and classes",
                      command=self.__show_model_head_class).pack(anchor="e", pady = 10)
        
        self.error_label = ctk.CTkLabel(master=self,  anchor='w', text=" ",
                                        fg_color='transparent', text_color = "red")
        self.error_label.pack(anchor="w", padx = 10)
        
        # ----------------------------------------------------------2. Model heads
        self.intro_label = ctk.CTkLabel(self.tabview.tab("2. Model head"), 
                                        text="1. Select model head")
        self.intro_label.grid(row=0, column=0, padx = 10, pady=(5,5), sticky="w")
        
        self.model_head_type =\
            ctk.CTkOptionMenu(self.tabview.tab("2. Model head"),
                              values=["Regression",
                                      "Gaussian Mixture Model (GMM)"],
                              command=self.__get_model_head_name)
        
        self.model_head_type.grid(row=1, column=0, padx = 10, pady=5, sticky="w")        
        CTkToolTip(self.model_head_type, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=250, justify = 'left',
                   message='Output from the model class (e.g., LSTM or EA-LSTM) ' +
                   'is input to the model head. Here model head could be Regression or GMM. ' +
                   'Regression is multi Linear layers (see PyTorch for nn.Linear) with user ' +
                   'choice activation function.')
        
        # --------------------------------------Frame for regression model head
        self.regression_frame = ctk.CTkFrame(master=self.tabview.tab("2. Model head"), 
                                             fg_color = "transparent", border_width=0.0)
        self.regression_frame.grid_columnconfigure((0,1), weight=1)
        self.regression_frame.grid(row=2, column=0, sticky="w", pady=(5, 5))
                
        # Number of layers
        self.regression_nlayers_label = ctk.CTkLabel(self.regression_frame,
                                  text="2. Number of layers")
        self.regression_nlayers_label.grid(row=0, column=0, sticky="w", padx=10, pady=5)
        
        self.regression_nlayers = ctk.CTkOptionMenu(self.regression_frame,
                          values=list(map(str, range(1, 6))),
                          command=self.__create_entry_regression_model)
        self.regression_nlayers.grid(row=1, column=0, sticky="w", padx=10, pady=5)        
        CTkToolTip(self.regression_nlayers, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message='Select number of linear layers. 1 should works for ' +
                   'most of the problems')
        
        # Label
        
        self.regression_config_label1 = ctk.CTkLabel(self.regression_frame,
                                  text="3. Number of neurons layer ith")
        self.regression_config_label1.grid(row=2, column=0, sticky="w", padx=10, pady=5) 
        CTkToolTip(self.regression_config_label1, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor='w',  wraplength=250, justify = "left",
                   message='The last layer is the output layer. Therefore, the number of '+
                   'neurons for this layer is the number of target features, users '+
                   'DO NOT need to input the number of neurons for the last layer ' +
                   'If there is only 1 layer, this layer is also the last layer ') 
        
        self.regression_config_label2 = ctk.CTkLabel(self.regression_frame,
                                  text="4. Activation function layer ith")
        self.regression_config_label2.grid(row=2, column=1, sticky="w", padx=10, pady=5)

        self.option_menu = ctk.CTkOptionMenu(self.regression_frame,
                                             values=["Identity", "ReLu", "Sigmoid",
                                                     "Tanh", "Softplus"],
                                             command=self.__reg_get_acts_1)
        self.option_menu.grid(row=3, column=1, sticky="e", padx=10, pady=5)
        self.reg_activation_func = [self.option_menu]
                
        # --------------------------------------------------------Frame for GMM
        self.gmm_frame = ctk.CTkFrame(master=self.tabview.tab("2. Model head"), 
                                             fg_color = "transparent", border_width=0.0)
        self.gmm_frame.grid_columnconfigure((0,1), weight=1)
        #self.gmm_frame.grid(row=2, column=0, sticky="w", pady=(5, 5))
                
        # Number of neuron in hidden layer
        self.gmm_hidden_size_label = ctk.CTkLabel(self.gmm_frame,
                                  text="2. Work in progress (comming soon)...")
        self.gmm_hidden_size_label.grid(row=0, column=0, sticky="w", padx=10, pady=5)
        self.gmm_hidden_size = ctk.CTkEntry(master=self.gmm_frame,
                                            placeholder_text="None")
        self.gmm_hidden_size.grid(row=1, column=0, sticky="w", padx=10, pady=5)
                  
        # ---------------------------------------------Content of load data tab
        self.intro_label = ctk.CTkLabel(self.tabview.tab("1. Model class"), 
                                                   text="1. Select model class")
        self.intro_label.pack(anchor="w", pady = (10,10))
        
        self.model_class_type =\
            ctk.CTkOptionMenu(self.tabview.tab("1. Model class"),
                              values=["LSTM","EA-LSTM"], command=self.__get_model_class)
        CTkToolTip(self.model_class_type, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message='The flow of data from input => output is as follows: ' + 
                   'Input => Model class => Model head => Ouput. ' +
                   'Model classes are: LSTM and EA-LSTM. Traditional LSTM, ' +
                   'dynamic and static inputs are concanated and feed into the LSTM. ' +
                   'while in EA-LSTM static input is feed into the input gate only. ' +
                   'See more in here https://doi.org/10.5194/hess-23-5089-2019')
                
        self.model_class_type.pack(anchor="w",  pady = 5)
                
        self.hidden_size_label = ctk.CTkLabel(self.tabview.tab("1. Model class"), 
                                                   text="2. Number of hidden units of the LSTM layer")
        self.hidden_size_label.pack(anchor="w", pady = (4,4))
        
        self.hidden_size = ctk.CTkEntry(master=self.tabview.tab("1. Model class"),
                                        placeholder_text="30")
        self.hidden_size.pack(anchor="w", pady = (4,4))
        CTkToolTip(self.hidden_size, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message='Input number of hidden neurons of the LSTM (or EA-LSTM) ' +
                   '(most problems should work with any values between 20 to 516). ' +
                   self.ray_tooltip)
        
        self.hidden_size.bind('<KeyRelease>', self.__get_hidden_size)

        self.nlayers_label = ctk.CTkLabel(self.tabview.tab("1. Model class"), 
                                               text="3. Number of LSTM layers")
        self.nlayers_label.pack(anchor="w", pady = (4,4))
        self.nlayers= ctk.CTkOptionMenu(self.tabview.tab("1. Model class"),
                                             values=list(map(str,list(range(1,6,1)))),
                                             command=self.__get_num_layers) 
        self.nlayers.pack(anchor="w", pady = (4,4))
        CTkToolTip(self.nlayers, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message='The number of layers of the LSTM (or EA-LSTM).' +
                   ' 1 should work for most of the problems')
        
        self.dropout_label = ctk.CTkLabel(self.tabview.tab("1. Model class"), 
                                               text="4. Dropout rate")
        self.dropout_label.pack(anchor="w", pady = (4,4))
        self.dropout = ctk.CTkEntry(master=self.tabview.tab("1. Model class"),
                                        placeholder_text="0.30")
        self.dropout.pack(anchor="w", pady = (4,4))
        self.dropout.bind('<KeyRelease>', self.__get_dropout)
        CTkToolTip(self.dropout, delay=0.1, bg_color = 'orange', justify = "left",
                   text_color = 'black', anchor='w',  wraplength=250, 
                   message='This value is only used if the number of layers > 1.' +
                   'During training, randomly zeroes some of the elements of the ' +
                   'input tensor to the i_th layer with probability p using samples ' +
                   'from a Bernoulli distribution. Commons values are from 0.2 to 0.5')

        
    # -------------------------------------------------content of load data tab
    def __show_model_head_class(self):
        
        # Get link to the image
        image = Path(__file__).parents[1]
        button_image = ctk.CTkImage(Image.open( 
            Path(image, "images/model_head_class.png")),
            size=(500, 500))
        
        # Create new window 
        self.new_window = ctk.CTkToplevel(self, fg_color = "white")
        ctk.CTkLabel(master=self.new_window, image=button_image, text = "").pack()
            
    def __get_hidden_size(self, dummy):
        # Get number of hidden layers
        get_input_text = self.hidden_size.get().strip()
        
        try:
            self.config["hidden_size"] = int(float(get_input_text))
            print("hidden size = ", self.config["hidden_size"])
            self.hidden_size.configure(fg_color = '#d3ffc7')
            self.error_label.configure(text=" ")
        except:
            try:
                eval(get_input_text)
                self.config["hidden_size"] = get_input_text
                print("hidden size = ", self.config["hidden_size"])
                self.hidden_size.configure(fg_color = '#d3ffc7')
                self.error_label.configure(text=" ")
            except:
                self.hidden_size.configure(fg_color = '#ffc7c7')
                self.error_label.configure(text=self.__message_text_tune("integer"), 
                                           text_color = "red")
            
    # Get number of lstm layers
    def __get_num_layers(self, nlayer: str):
        try:
            self.config["num_layers"] = int(nlayer)
            print("num_layers = ", self.config["num_layers"])
        except:
            pass
        
    # Message box for error in input with tune search space 
    def __message_text_tune(self, data_type):
        out_message=("Input should be " + data_type + " or Tune Search Space API command: " +
            "https://docs.ray.io/en/latest/tune/api/search_space.html")
        return out_message

            
    # Get number of lstm layers
    def __get_dropout(self, dummy: str):
        
        get_input_text = self.dropout.get().strip()
        
        try:
            self.config["dropout"] = float(get_input_text)
            print("drop out rate = ", self.config["dropout"])
            self.dropout.configure(fg_color = '#d3ffc7')
        except:
            try:
                eval(get_input_text)
                self.config["dropout"] = get_input_text
                print("drop out rate = ", self.config["dropout"])
                self.dropout.configure(fg_color = '#d3ffc7')
            except:
                self.dropout.configure(fg_color = '#ffc7c7')

    def __get_model_head_name(self, model_head_name: str):
        model_head_names = {"Regression" : "Regression",
                            "Gaussian Mixture Model (GMM)": "GMM"}
        
        self.globalData["model_head"] = model_head_names[model_head_name]
        print("Model head name = ", self.globalData["model_head"])
            
        # Delete config if other model head option is 
        if self.globalData["model_head"] == "Regression":
                       
            self.regression_frame.grid(row=2, column=0, sticky="w", pady=(5, 5))
            self.gmm_frame.grid_forget()
            
            self.config["Regression"] = {}
        else:
            # Hide regression frame
            self.regression_frame.grid_forget()
            self.gmm_frame.grid(row=2, column=0, sticky="w", pady=(5, 5))
            try:
                del self.config["Regression"]
            except:
                pass
 
    def __get_model_class(self, model_class: str):
        self.config["model_class"] = model_class
        print("Model class name = ", self.config["model_class"])
        
        if model_class == "EA-LSTM":
            tk.messagebox.showinfo(title="Message box", 
                                   message="This model class takes a lot of time to train " +
                                   "need to improve the forward pass for this model class")         
            
    def __create_entry_regression_model(self, nlayers: str):
        print("number of layer", int(nlayers))
        
        # try to hide all entry
        try:
            for entry in self.reg_neurons:
                entry.grid_forget()
        except:
            pass
        
        # try to hide all entry
        try:
            for entry in self.reg_activation_func:
                entry.grid_forget()
        except:
            pass
        
        # Now add entry according to the number of layers
        if "Regression" in self.config.keys():
            
            row_number = 3
            self.config["Regression"]["num_layers"] = int(nlayers)
            self.reg_neurons = []
            self.reg_activation_func = []
            self.config["Regression"]["num_neurons"] = [None for i in range(int(nlayers))]
            self.config["Regression"]["activation_function"] = ["Identity" for i in 
                                                         range(int(nlayers))]

            get_neurons = {0: self.__reg_get_neurons_1, 1: self.__reg_get_neurons_2,
                           2: self.__reg_get_neurons_3, 3: self.__reg_get_neurons_4}
            get_acts = {0: self.__reg_get_acts_1, 1: self.__reg_get_acts_2,
                        2: self.__reg_get_acts_3, 3: self.__reg_get_acts_4,
                        4: self.__reg_get_acts_5}
            
            for i in range(int(nlayers)):

                if i < int(nlayers) - 1:
                    entry = ctk.CTkEntry(master=self.regression_frame,
                                         placeholder_text="10")
                    entry.grid(row=row_number, column=0, sticky="e", padx=10, pady=5)
                    self.reg_neurons.append(entry)
                    self.reg_neurons[i].bind('<KeyRelease>', get_neurons[i])
                
                option_menu = ctk.CTkOptionMenu(self.regression_frame,
                                                     values=["Identity", "ReLu", "Sigmoid",
                                                             "Tanh", "Softplus"],
                                                     command=get_acts[i])
                option_menu.grid(row=row_number, column=1, sticky="e", padx=10, pady=5)
                self.reg_activation_func.append(option_menu)
                row_number += 1
                
                try:
                    CTkToolTip(self.reg_neurons[i], delay=0.1, bg_color = 'orange',
                               text_color = 'black', anchor='w',  wraplength=250, justify = "left",
                               message='Input the number of neurons of the ' + 
                               str(i + 1) + ' layer (please give an integer number)')
                except:
                    pass
                
    # Function get number of neuron in each output layer
    def __get_num_neuron(self, layer):    
        try:
            get_input_text = self.reg_neurons[layer].get().strip()
            self.config["Regression"]["num_neurons"][layer] = int(float(get_input_text))
            print("Num of neurons layer " + str(layer + 1) + " = ", 
                  self.config["Regression"]["num_neurons"][layer])
            self.reg_neurons[layer].configure(fg_color = '#d3ffc7')
        except:
            self.reg_neurons[layer].configure(fg_color = '#ffc7c7')


    # Functions to get number of input neurons each layer - Make this code clearner
    def __reg_get_neurons_1(self, dummy):
        self.__get_num_neuron(0)

    def __reg_get_neurons_2(self, dummy):
        self.__get_num_neuron(1)

    def __reg_get_neurons_3(self, dummy):
        self.__get_num_neuron(2)
        
    def __reg_get_neurons_4(self, dummy):
        self.__get_num_neuron(3)
    
    # Function to get activation function of each layer - Make this code clearner
    def __reg_get_acts_1(self, dummy):
        self.config["Regression"]["activation_function"][0] = dummy
        print("Activation function of layer 1 = ", dummy)
    def __reg_get_acts_2(self, dummy):
        self.config["Regression"]["activation_function"][1] = dummy
        print("Activation function of layer 2 = ", dummy)
    def __reg_get_acts_3(self, dummy):
        self.config["Regression"]["activation_function"][2] = dummy
        print("Activation function of layer 3 = ", dummy)
    def __reg_get_acts_4(self, dummy):
        self.config["Regression"]["activation_function"][3] = dummy
        print("Activation function of layer 4 = ", dummy)
    def __reg_get_acts_5(self, dummy):
        self.config["Regression"]["activation_function"][5] = dummy
        print("Activation function of layer 5 = ", dummy)




import customtkinter as ctk
from pathlib import Path
 
from hydroecolstm.interface.data_frame import DataFrame
from hydroecolstm.interface.network_design_frame import NetworkDesignFrame
from hydroecolstm.interface.project_summary_frame import ProjectSummaryFrame
from hydroecolstm.interface.sidebar_frame import SidebarFrame
from hydroecolstm.interface.train_test_frame import TrainTestFrame
from hydroecolstm.interface.visualize_frame import VisualizeFrame
from hydroecolstm.interface.application_frame import ApplicationFrame



class MainGUI(ctk.CTk):
    def __init__(self):

        # Initialize the interface apprearence
        ctk.set_appearance_mode("Light")
        ctk.set_default_color_theme("dark-blue")
        ctk.set_widget_scaling(1.1)

        # Initialize project setting - will be replaced by user defined values in GUI
        global config
        global globalData
        
        config = {}
        config["scaler_input_dynamic_features"] = ["MinMaxScaler"]
        config["scaler_input_static_features"] = ["MinMaxScaler"]
        config["scaler_target_features"] = ["MinMaxScaler"]
        config["hidden_size"] = 30
        config["num_layers"] = 1
        config["activation_function_name"] = "Identity"
        config["n_epochs"] = 50
        config["learning_rate"] = 0.01
        config["dropout"] = 0.30
        config["warmup_length"] = 20      
        config["sequence_length"] = 720
        config["batch_size"] = 3
        config["patience"] = 20
        config["loss_function"] = "RMSE"
        config["Regression"] = {}
        config["Regression"]["activation_function"] = ["Identity"]
        config["Regression"]["num_neurons"] = [None]
        config["Regression"]["num_layers"] = 1
        config["hidden_size"] = 30
        config["dropout"] = 0.30
        config["n_epochs"] = 5
        config["learning_rate"] = 0.01
        config["warmup_length"] = 20
        config["model_class"] = "LSTM"
        config["output_directory"] = [Path.cwd()]

        # Initialize global data
        globalData = {}
        globalData['dynamic_data_header'] =[]
        globalData['init_state_dicts'] = False
        globalData['static_data_header'] =[]
        globalData["object_id_no"] = 0
        globalData["target_feature_no"] = 0
        globalData["object_id_forecast_no"] = 0
        globalData["target_feature_forecast_no"] = 0
        globalData["model_head"] = "Regression"
        
        super().__init__()
    
        #------------------------------------------------------configure window
        self.title("Hydro-ecological modelling with Long Short-Term Memory neural network")
        self.geometry(f"{1100}x{580}")
    
        #--------------------------------------------------------configure grid
        self.grid_columnconfigure(0, weight=0) 
        self.grid_columnconfigure(1, weight=1)
        self.grid_columnconfigure(2, weight=1)
        self.grid_columnconfigure(3, weight=0)
        self.grid_rowconfigure((0,1,2,3), weight=1)
    
        #----------------------------------------------------create main frames
        # data frame
        self.data_frame = DataFrame(container=self, config=config, 
                                    globalData=globalData)
        self.data_frame.grid(row=0, column=1, rowspan=4, columnspan=2,
                             padx = 10, pady=(20, 20), sticky="nsew")
    
        # network frame
        self.network_frame = NetworkDesignFrame(container=self, config=config,
                                                globalData=globalData)
            
        # train test frame
        self.train_test_frame = TrainTestFrame(container=self, config=config,
                                               globalData=globalData)
    
        # visualize frame
        self.visual_frame = VisualizeFrame(container=self, config=config,
                                           globalData=globalData)

        # visualize frame
        self.application_frame = ApplicationFrame(container=self, config=config,
                                           globalData=globalData)
            
        # summary frame
        self.summary_frame = ProjectSummaryFrame(container=self, config=config,
                                                 globalData=globalData)
        self.summary_frame.grid(row=0, column=3, rowspan=4, padx = 10,
                                pady=(20, 20), sticky="nsew") 
            
        # sidebar frame - main frame controlling other frame
        self.sidebar_frame = SidebarFrame(container=self, 
                                          data_frame=self.data_frame,
                                          network_frame=self.network_frame,
                                          train_test_frame=self.train_test_frame,
                                          visual_frame=self.visual_frame,
                                          application_frame=self.application_frame)
        
        self.sidebar_frame.grid(row=0, column=0, rowspan=4, padx = 0,
                                pady=(20, 20), sticky="nsew") 

                 
# Display GUI 
def show_gui():
    app = MainGUI()
    app.mainloop()  


import matplotlib
import customtkinter as ctk
import tkinter as tk
import torch
from pandastable import Table
from hydroecolstm.data.read_config import read_config
from hydroecolstm.interface.utility import (ToplevelWindow, 
                                            plot_train_valid_loss,
                                            plot_time_series,
                                            check_size,check_linestyle,
                                            check_alpha,check_line_plot,
                                            check_marker, check_color, 
                                            check_ylim,
                                            combine_simulated)
from CTkToolTip import CTkToolTip

    
class VisualizeFrame(ctk.CTkScrollableFrame):
    def __init__(self, container=None, config=None, globalData=None):
        super().__init__(container)
        
        # setup the grid layout manager
        self.config=config
        self.globalData=globalData
        self.columnconfigure((0), weight=1)
        self.__create_widgets() 
        
        self.plot_window=None
        
    # create widgets for sidebar frame
    def __create_widgets(self):
        # create tabs

        self.tabview=ctk.CTkTabview(self, width=750, border_width=1.5,
                                      fg_color="transparent")
        self.tabview.grid(row=0, column=0, sticky="we")
        self.tabview.add("1. Data for plot")
        self.tabview.tab("1. Data for plot").grid_columnconfigure((0,1),weight=1)
        self.tabview.add("2. Loss plot")
        self.tabview.tab("2. Loss plot").grid_columnconfigure((0,1,2), weight=1)
        self.tabview.add("3. Time-series plot")
        self.tabview.tab("3. Time-series plot").grid_columnconfigure((0,1,2), weight=1)
        #----------------------------------------------------------------------
        self.data_plot_label=ctk.CTkLabel(self.tabview.tab("1. Data for plot"),
                                         text="1. Select data for plot")
        self.data_plot_label.grid(row=0, column=0, sticky="w", padx=5, pady=5)
        
        self.data_selection = ctk.CTkOptionMenu(self.tabview.tab("1. Data for plot"),
                                                values=['from current project',
                                                        'from other project'],
                                                command=self.data_selection_event)
        self.data_selection.grid(row=1, column=0, sticky="w", padx=5, pady=5)
        CTkToolTip(self.data_selection, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message="Select 'from other project' ONLY IF you want to use "+ 
                   "data from other project")
        
        #----------------------------------------------
        self.loss_label=ctk.CTkLabel(self.tabview.tab("2. Loss plot"),
                                            text="1. Plot training and validation loss")
        self.loss_label.grid(row=0, column=0, columnspan=3, sticky="w", padx=5, pady=5)
        
        self.loss_plot=ctk.CTkButton(self.tabview.tab("2. Loss plot"), anchor='w', 
                                         command=self.loss_plot_event,
                                         text="Plot (update plot)")
        self.loss_plot.grid(row=1, column=0, sticky="w", padx=5, pady=5)
        CTkToolTip(self.loss_plot, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message='Click here to create plot or update plot (in a new window)')
        
        # check to load/unload
        self.plot_loss=ctk.IntVar(value=0)        
        self.plot_loss_checkbox=ctk.CTkCheckBox(self.tabview.tab("2. Loss plot"), 
                                               text="Show plot settings",
                                               command=self.show_plot_loss_setting, 
                                               variable=self.plot_loss)
        self.plot_loss_checkbox.grid(row=2, column=0, padx=10, pady=(10, 0), 
                                     sticky="w")
        CTkToolTip(self.plot_loss_checkbox, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='Change plot settings here or ' + 
                   'leave it empty for default settings')
        
        self.plot_loss_frame=ctk.CTkFrame(self.tabview.tab("2. Loss plot"),
                                            height=400, fg_color="transparent")
        self.plot_loss_frame.columnconfigure((0,1,2), weight=1)
        
        #-----------------------------------------------------Plot loss setting 
        self.common_settings_loss_plot=ctk.CTkLabel(self.plot_loss_frame,
                                         text="Common settings")
        self.common_settings_loss_plot.grid(row=0, column=0, sticky="w", 
                                            padx=5, pady=5)
        
        # Plot title
        self.title=ctk.CTkEntry(self.plot_loss_frame,
                                  placeholder_text="plot title")
        self.title.grid(row=1, column=0, sticky="w", padx=5, pady=5)
        
        # X label
        self.xlabel=ctk.CTkEntry(self.plot_loss_frame, 
                                   placeholder_text="xlabel")
        self.xlabel.grid(row=2, column=0, sticky="w", padx=5, pady=5)

        # Y label
        self.ylabel=ctk.CTkEntry(self.plot_loss_frame, 
                                   placeholder_text="ylabel")
        self.ylabel.grid(row=3, column=0, sticky="w", padx=5, pady=5)

        # Legend name of the best model
        self.best_model_legend=ctk.CTkEntry(self.plot_loss_frame , 
                                  placeholder_text="best model name")
        self.best_model_legend.grid(row=4, column=0, sticky="w", 
                                    padx=5, pady=5)
        
        #---------------------------------------------------------Training line
        # training line
        self.training_line=ctk.CTkLabel(self.plot_loss_frame,
                                         text="Training loss line")
        self.training_line.grid(row=0, column=1, sticky="w", padx=5, pady=5)
        
        # Color of the training loss line
        self.train_line_color=ctk.CTkEntry(self.plot_loss_frame , 
                                  placeholder_text="color")
        CTkToolTip(self.train_line_color, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e', wraplength=500, justify="left", 
                   message=str(list(matplotlib.colors.cnames.keys())))
        
        self.train_line_color.grid(row=1, column=1, sticky="w", padx=5, pady=5)

        # Line type of the training loss line
        self.train_line_style=ctk.CTkEntry(self.plot_loss_frame , 
                                  placeholder_text="line stype")
        CTkToolTip(self.train_line_style, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='solid, dashed, dashdot, dotted')
        
        self.train_line_style.grid(row=2, column=1, sticky="w", padx=5, pady=5)

        # Line type of the training loss line
        self.train_legend=ctk.CTkEntry(self.plot_loss_frame, 
                                  placeholder_text="legend name")
        self.train_legend.grid(row=3, column=1, sticky="w", padx=5, pady=5)
              
        #-------------------------------------------------------Validation line
        self.validation_line=ctk.CTkLabel(self.plot_loss_frame,
                                         text="Validation loss line")
        self.validation_line.grid(row=0, column=2, sticky="w", padx=5, pady=5)
        
        # Color of the validation loss line
        self.valid_line_color=ctk.CTkEntry(self.plot_loss_frame , 
                                  placeholder_text="color")
        self.valid_line_color.grid(row=1, column=2, sticky="w", padx=5, pady=5)
        CTkToolTip(self.valid_line_color, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e', wraplength=500, justify="left", 
                   message=str(list(matplotlib.colors.cnames.keys())))
        
        # Line type of the validation loss line
        self.valid_line_style=ctk.CTkEntry(self.plot_loss_frame , 
                                  placeholder_text="line style")
        self.valid_line_style.grid(row=2, column=2, sticky="w", padx=5, pady=5)
        CTkToolTip(self.valid_line_style, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='solid, dashed, dashdot, dotted')

        # Line type of the training loss line
        self.valid_legend=ctk.CTkEntry(self.plot_loss_frame , 
                                  placeholder_text="legend name")
        CTkToolTip(self.valid_legend, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message='Validation line legend: type name of the validation line here, \n' +
                   'or leave it empty to use default name')
        self.valid_legend.grid(row=3, column=2, sticky="w", padx=5, pady=5)
        
        # ---------------------------------------------content of load data tab
        self.object_id_label =\
            ctk.CTkLabel(self.tabview.tab("3. Time-series plot"),
                         text="1. Select object id and target feature for plot")
            
        self.object_id_label.grid(row=0, column=0, columnspan=3, sticky="w", padx=5)

        self.object_id=ctk.CTkTextbox(self.tabview.tab("3. Time-series plot"), 
                                        height=30, border_width=1.5)
        
        self.object_id.insert("0.0", "object_id") 
        self.object_id.grid(row=1, column=0, sticky="w", padx=5, pady=5)
        
        self.object_id.bind('<KeyRelease>', self.get_object_id)

        self.target_feature=ctk.CTkTextbox(self.tabview.tab("3. Time-series plot"), 
                                             height=30, border_width=1.5)
        
        self.target_feature.insert("0.0", "target_feature") 
        
        self.target_feature.grid(row=2, column=0, sticky="w", padx=5, pady=5)
        
        self.target_feature.bind('<KeyRelease>', self.get_target_feature)
        
        self.next_object_id_button=ctk.CTkButton(self.tabview.tab("3. Time-series plot"), 
                                                   anchor='w',
                                                   command=self.next_object_id, 
                                                   text="Next object id")
        
        self.next_object_id_button.grid(row=1, column=1, sticky="w", 
                                        padx=5, pady=5)
        
        self.next_target_feature_button =\
            ctk.CTkButton(self.tabview.tab("3. Time-series plot"), anchor='w',
                          command=self.next_target_feature,
                          text="Next target features")
            
        self.next_target_feature_button.grid(row=2, column=1, sticky="w", 
                                             padx=5, pady=5)
        
        # Plot button
        self.update_plot=ctk.CTkButton(self.tabview.tab("3. Time-series plot"), anchor='we', 
                                 command=self.plot_figure, text="Plot (update plot)")
        
        self.update_plot.grid(row=1, column=2,  
                              sticky="we", padx=5, pady=5)
        
        # show all data
        self.show_all_data=ctk.CTkButton(self.tabview.tab("3. Time-series plot"), anchor='w', 
                                         command=self.show_all_data_event,
                                         text="Show data source")
        self.show_all_data.grid(row=2, column=2, sticky="we", padx=5, pady=5)
        CTkToolTip(self.show_all_data, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message='Click here to show all simulated and observed data')
        
        
        #-------------------------------------------------------2. Plot setting
        # check to load/unload
        self.plot_timeseries=ctk.IntVar(value=0)        
        self.plot_timeseries_checkbox=ctk.CTkCheckBox(self.tabview.tab("3. Time-series plot"), 
                                               text="Show plot settings",
                                               command=self.show_plot_timeseries_setting, 
                                               variable=self.plot_timeseries)
        
        self.plot_timeseries_checkbox.grid(row=3, column=0, padx=10, pady=(10, 0), sticky="w")
        
        self.plot_timeseries_frame=ctk.CTkFrame(self.tabview.tab("3. Time-series plot"),
                                              height=400, fg_color="transparent")
        self.plot_timeseries_frame.columnconfigure((0,1,2), weight=1)

        #----------------------------------------------------------------------
        self.common_settings=ctk.CTkLabel(self.plot_timeseries_frame,
                                         text="Common settings")   
        self.common_settings.grid(row=0, column=0, sticky="w", padx=5)
        
        self.plot_title=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="plot title")
        
        self.plot_title.grid(row=1, column=0, sticky="w", padx=5)
        
        
        self.x_label=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="xlabel")
        self.x_label.grid(row=2, column=0, sticky="w", padx=5)
        
        self.y_label=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="ylabel")
        self.y_label.grid(row=3, column=0, sticky="w", padx=5)
        
        self.ylim_up=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="ylim (upper)")
        CTkToolTip(self.ylim_up, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left",
                   message='Upper limit of the y axis (input numeric value)')
        
        self.ylim_up.grid(row=4, column=0, sticky="w", padx=5)
        
        self.observed_data_plot=ctk.CTkLabel(self.plot_timeseries_frame,
                                         text="observed data plot")
        self.observed_data_plot.grid(row=0, column=1, sticky="w", padx=5)
          
        self.lineplot=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="line (true or false)")
        self.lineplot.grid(row=1, column=1, sticky="w", padx=5, pady=5)
                
        self.color_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="color")
        self.color_obs.grid(row=2, column=1, sticky="w", padx=5, pady=5)
        CTkToolTip(self.color_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", 
                   message=str(list(matplotlib.colors.cnames.keys())))
        
        self.alpha_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="alpha")    
        self.alpha_obs.grid(row=3, column=1, sticky="w", padx=5, pady=5)
        CTkToolTip(self.alpha_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='tranparency level: input any ' +
                   'numeric values between 0 and 1')
        
        self.size_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="size")        
        self.size_obs.grid(row=4, column=1, sticky="w", padx=5, pady=5)
        CTkToolTip(self.size_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message='input positive numeric value for line thickness')
        
        self.linestyle_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="linestyle")    
        CTkToolTip(self.linestyle_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='solid, dashed, dashdot, dotted')
        
        self.linestyle_obs.grid(row=5, column=1, sticky="w", padx=5, pady=5)
        self.marker_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                       placeholder_text="marker")
        self.marker_obs.grid(row=6, column=1, sticky="w", padx=5, pady=5)
        CTkToolTip(self.marker_obs, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message="., o, s, ^, v, +, x")
        
        self.label_obs=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="label")
        
        self.label_obs.grid(row=7, column=1, sticky="w", padx=5, pady=5)
        
        #----------------------------------------------------------------------
        self.simulated_data_plot=ctk.CTkLabel(self.plot_timeseries_frame,
                                         text="simulated data plot")
        self.simulated_data_plot.grid(row=0, column=3, sticky="w", padx=5)
        
        self.color_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="color")
        self.color_sim.grid(row=2, column=3, sticky="w", padx=5)
        self.alpha_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="alpha")
        self.alpha_sim.grid(row=3, column=3, sticky="w", padx=5)
        self.size_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="size")
        self.size_sim.grid(row=4, column=3, sticky="w", padx=5)
        CTkToolTip(self.size_sim, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, justify="left", 
                   message='input positive numeric value for line thickness')
        
        self.linestyle_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="linestype")
        self.linestyle_sim.grid(row=5, column=3, sticky="w", padx=5)
        CTkToolTip(self.linestyle_sim, delay=0.1, bg_color='orange',
                   text_color='black', anchor='e',  wraplength=500, 
                   justify="left", message='solid, dashed, dashdot, dotted')
        self.label_sim=ctk.CTkEntry(self.plot_timeseries_frame,
                                  placeholder_text="label")
        self.label_sim.grid(row=7, column=3, sticky="w", padx=5)

    def show_all_data_event(self):
        try:
            dataframe = combine_simulated(self.globalData, 
                                          self.globalData["y_column_name"])
            
            self.table = Table(tk.Toplevel(self), dataframe=dataframe, 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()
        except:
            pass
        
        
    # Data selection event
    def data_selection_event(self, method: str):
        if method == "from other project":
            #  Load config file
            self.load_data_label=ctk.CTkLabel(self.tabview.tab("1. Data for plot"),
                                             text="2. Load data from other projects (ATTENTION!!!)")
            self.load_data_label.grid(row=2, column=0, padx=5, pady=5, 
                                      sticky="w")
            self.load_config=ctk.CTkButton(self.tabview.tab("1. Data for plot"),
                                           text="Select config.yml file type",
                                           command=self.load_config_event)
            self.load_config.grid(row=3, column=0, padx=5, pady=5, sticky="w")
            CTkToolTip(self.load_config, delay=0.1, bg_color='orange',
                       text_color='black', anchor='e',  wraplength=500, justify="left", 
                       message='Click this button to load config.yml from other '+
                       'project and OVERWRITE configuration created from previous steps ' +
                       '(if they exist)')
            
            self.selected_config_file = ctk.CTkLabel(self.tabview.tab("1. Data for plot"),
                                                          text="No file was selected")
            self.selected_config_file.grid(row=3, column=1, padx=5, pady=5, sticky="w")             
 
    
            self.load_data=ctk.CTkButton(self.tabview.tab("1. Data for plot"),
                                           text="Select data.pt file type",
                                           command=self.load_data_event)
            self.load_data.grid(row=4, column=0, padx=5, pady=5, sticky="w") 
            CTkToolTip(self.load_config, delay=0.1, bg_color='orange',
                       text_color='black', anchor='e',  wraplength=500, justify="left", 
                       message='Click this button to load data.pt from other '+
                       'project and OVERWRITE data created from previous steps ' +
                       '(if they exist)')
            self.selected_data_file = ctk.CTkLabel(self.tabview.tab("1. Data for plot"),
                                                          text="No file was selected")
            self.selected_data_file.grid(row=4, column=1, padx=5, pady=5, sticky="w") 
        else:
            try:
                # hide all buttons for load data from previous project
                self.load_data_label.grid_forget()
                self.load_config.grid_forget()
                self.selected_config_file.grid_forget()
                self.load_data.grid_forget()
                self.selected_data_file.grid_forget()
            except:
                pass
            
    
    def load_config_event(self):
        try:
            file_name=ctk.filedialog.askopenfilename(title="Select config.yml file type", 
                                                     filetypes=(('yml files', '*.yml'),
                                                                ('All files', '*.*')))
            self.config.update(read_config(file_name))
            self.selected_config_file.configure(text= '...' + file_name[-30:])
        except:
            pass
        
    def load_data_event(self):
        try:
            file_name=ctk.filedialog.askopenfilename(title="Select data.pt file type", 
                                                     filetypes=(('pt files', '*.pt'),
                                                                ('All files', '*.*')))
            self.globalData.update(torch.load(file_name))
            self.selected_data_file.configure(text= '...' + file_name[-30:])
            print(self.globalData.keys())
        except:
            pass
        
        
    # Show plot loss setting
    def show_plot_loss_setting(self):
        if self.plot_loss.get() == 0:
            self.plot_loss_frame.grid_forget()
        else:
            self.plot_loss_frame.grid(row=3, column=0, columnspan=3,
                                      sticky="w", padx=(0,0), pady=(20,20))

    # Show plot time series setting
    def show_plot_timeseries_setting(self):
        if self.plot_timeseries.get() == 0:
            self.plot_timeseries_frame.grid_forget()
        else:
            self.plot_timeseries_frame.grid(row=4, column=0, columnspan=3,
                                      sticky="w", padx=(0,0), pady=(20,20))
            
    # Get next object id
    def next_object_id(self):
        try:
            if self.globalData["object_id_no"] > len(self.config["object_id"]) - 1:
                self.globalData["object_id_no"]=0
            
            self.object_id.delete("0.0", "end")
            
            self.object_id.insert("0.0", self.config["object_id"]
                                  [self.globalData["object_id_no"]])
            
            self.globalData["object_id_plot"] =\
                str(self.config["object_id"][self.globalData["object_id_no"]])
                
            self.globalData["object_id_no"] += 1
        except:
            pass

    def next_target_feature(self):
        try:
            if self.globalData["target_feature_no"] > len(self.config["target_features"]) - 1:
                self.globalData["target_feature_no"]=0
               
            self.target_feature.delete("0.0", "end")
            self.target_feature.insert("0.0", self.config["target_features"]
                                       [self.globalData["target_feature_no"]])
            self.globalData["target_feature_plot"] =\
                str(self.config["target_features"][self.globalData["target_feature_no"]])
                
            self.globalData["target_feature_no"] += 1
        except:
            pass
            
    def get_object_id(self, dummy):
        self.globalData["object_id_plot"]=str(self.object_id.get("0.0", "end"))
        self.globalData["object_id_plot"]=self.globalData["object_id_plot"].strip()
        print(f"Selected object_id for plot={self.globalData['object_id_plot']}")

    def get_target_feature(self, dummy):
        self.globalData["target_feature_plot"] =\
            str(self.target_feature.get("0.0", "end"))
            
        self.globalData["target_feature_plot"] =\
            self.globalData["target_feature_plot"].strip()
            
        print(f"Selected target_feature for plot={self.globalData['target_feature_plot']}")
       
    def plot_figure(self):
        try:
            # delete later
            key=self.globalData["object_id_plot"]
            idx=self.config["target_features"].index(self.globalData["target_feature_plot"])
            
            # General setting for plotting observed data
            lineplot=check_line_plot(self.lineplot)
            color_obs=check_color(self.color_obs, "coral")
            alpha_obs=check_alpha(self.alpha_obs, 1)
            size_obs=check_size(self.size_obs, 1)
            linestyle_obs=check_linestyle(self.linestyle_obs, 'dashed')
            marker_obs=check_marker(self.marker_obs, "d")
            label_obs=self.label_obs.get()
            if len(label_obs) == 0: label_obs="Observed"

            # General setting for plotting simulated data
            color_sim=check_color(self.color_sim, "blue")
            alpha_sim=check_alpha(self.alpha_sim, 1)
            size_sim=check_size(self.size_sim, 1)
            linestyle_sim=check_linestyle(self.linestyle_sim, 'solid')
            label_sim=self.label_sim.get()
            if len(label_sim) == 0: label_sim="Simulated"
          
            # Lable
            title=self.plot_title.get()
            xlabel=self.x_label.get()
            ylabel=self.y_label.get().strip()
            if ylabel == "": ylabel=self.globalData["target_feature_plot"]
            ylim_up=check_ylim(self.ylim_up, -9999.0)
                
            plot_window=ToplevelWindow(window_name="Plot window")
            
            # Plot
            plot_time_series(plot_window, self.globalData, key, idx, lineplot, 
                             color_obs, alpha_obs, size_obs, linestyle_obs,
                             marker_obs, label_obs, color_sim, alpha_sim, 
                             size_sim, linestyle_sim, label_sim, title, xlabel,
                             ylabel, ylim_up, forecast_period=False)
            
            
        except:
            tk.messagebox.showinfo(title="Error", 
                                   message="Cannot plot time series")
        
    def loss_plot_event(self):
        try:
            # Data for plot

            loss=self.globalData["loss_epoch"]

            # Make plot window
            plot_window=ToplevelWindow(window_name="Plot window")
            
            # Now plot
            plot_train_valid_loss(loss, 
                                  plot_window, 
                                  self.xlabel.get(),
                                  self.ylabel.get(), 
                                  self.title.get(),
                                  self.train_line_color.get().strip(),
                                  self.train_line_style.get().strip(),
                                  self.train_legend.get().strip(),
                                  self.valid_line_color.get().strip(),
                                  self.valid_line_style.get().strip(),
                                  self.valid_legend.get().strip(),
                                  self.best_model_legend.get().strip())
        except:
            # Make plot window
            #plot_window=ToplevelWindow(window_name="Plot window")
            #plot_time_series(plot_window)
            
            tk.messagebox.showinfo(title="Error", 
                                   message="Cannot plot loss")
            
    def show_loss_pandastable(self):
        try:
            self.table = Table(tk.Toplevel(self), dataframe=self.globalData['dynamic_data'], 
                               showtoolbar=True, showstatusbar=True)
            self.table.show()
        except:
            None





import customtkinter as ctk
import tkinter as tk
from CTkToolTip import CTkToolTip
from CTkMessagebox import CTkMessagebox
import torch
from pathlib import Path

from hydroecolstm.interface.utility import config_to_text, sort_key
from hydroecolstm.interface.utility import write_yml_file

class ProjectSummaryFrame(ctk.CTkFrame):
    def __init__(self, container=None, config=None, globalData=None):
        super().__init__(container)
        self.config = config
        self.globalData = globalData
        # setup the grid layout manager
        self.columnconfigure(0, weight=1)
        self.rowconfigure((0), weight=0)
        self.rowconfigure((1), weight=1)
        self.rowconfigure((2), weight=0)
        self.__create_widgets() 
        
    # create widgets for sidebar frame
    def __create_widgets(self):     
        self.update_summary = ctk.CTkButton(self, text="Project Summary", 
                                            font=ctk.CTkFont(size=20, weight="bold"),
                                            command=self.update_project_summary, 
                                            fg_color = "transparent",
                                            text_color="black")
        self.update_summary.grid(row=0, column=0, pady=0, padx=0)
        CTkToolTip(self.update_summary, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'w', 
                   message= 'Click here to update the project summary') 
        
        self.summary_textbox = ctk.CTkTextbox(master=self,corner_radius=0, 
                                              height=2000,
                                              bg_color='transparent', 
                                              fg_color='transparent',
                                              activate_scrollbars=True,
                                              wrap='none')

        self.summary_textbox.grid(row=1, column=0,pady=(10,7), padx=0)
        self.summary_textbox.insert("end", "Click 'Project Summary'\n" )
        self.summary_textbox.insert("end", "to see project info: " )
        self.summary_textbox.configure(spacing3=10) 

        self.save_buton = ctk.CTkButton(self, fg_color='gray',
                                        border_width=0,
                                        command=self.save_yml,
                                        text = "Save")
        CTkToolTip(self.save_buton, delay=0.1, bg_color = 'orange',
                   text_color = 'black', anchor = 'n',  wraplength=500, 
                   message= 'Click here to save project summary as' +
                   ' configuration file (here, you can give the file name)' +
                   ' or save all (all data created by this tool + the model +' +
                   ' configuration file (here you cannot give the file name,' +
                   ' just select the folder and files with predefined names will be saved')

        self.save_buton.grid(row=2, column=0,pady=(10,7), padx=0)
        
    def update_project_summary(self):
        # Delete text
        self.summary_textbox.delete("0.0", "end")
        output_text = config_to_text(config=sort_key(self.config))
    
        for text in output_text:
            self.summary_textbox.insert("end", text)
    

    def save_yml(self):
        
        # Ask which one user want to save
        msg = CTkMessagebox(title="Save", message="Please select save option",
                    option_1="Cancel", 
                    option_3="Save all",
                    option_2="Save config")
        response = msg.get()

        try:
            if response == "Save config":
                
                # Save config as .yml
                file_name = ctk.filedialog.asksaveasfilename(
                    title="Save project summary as .yml file",
                    filetypes=(('yml files', '*.yml'),
                               ('All files', '*.*')))
                write_yml_file(config=self.config, out_file=file_name)
                print("Saved project summary as ", file_name)
                
            elif response == "Save all":
                # Select dir to save      
                output_directory = tk.filedialog.askdirectory()

                # Save config as .yml                 
                write_yml_file(config=self.config, 
                               out_file=Path(output_directory, "config.yml"))
                print("Saved project summary as config.yml file")
                
                print(self.globalData.keys())
                if "best_config" in self.globalData.keys():
                    write_yml_file(config=self.config, 
                                   out_file=Path(output_directory, "config.yml"))
                    print("Saved trained config as best_config.yml file")                    
        
                # Save model_state_dicts to model_state_dict.pt file
                torch.save(self.globalData["model"].state_dict(),
                           Path(output_directory, "model_state_dict.pt"))
                
                print("Model state_dict was saved as model_state_dict.pt")
                
                # Save global data
                torch.save(self.globalData, Path(output_directory, "data.pt"))
                print("globalData was saved as data.pt")
            else:
                pass
                
        except:
            print("Error: Cannot save model or data")            
            
